#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext
from .. import log as _log


def mavenCentral(env):
    return env.get('mavenCentral',
                   'http://repo.maven.apache.org/maven2')


def ensureJDBC(env, lib_dir, drivers):
    ospath = _ext.ospath
    pathutil = _ext.pathutil
    ver = env.get('javaVer', '8')

    maven_central = mavenCentral(env)

    psql_ver = '42.1.3'

    if int(ver) == 7:
        psql_ver += '.jdk7'
    elif int(ver) == 6:
        psql_ver += '.jdk6'

    driverMap = {
        'mysql': ('mysql', 'mysql-connector-java', '5.1.43'),
        'postgresql': ('org/postgresql', 'postgresql', psql_ver),
        'mssql': ('com/microsoft/sqlserver', 'mssql-jdbc', '6.2.1.jre' + ver),
        'sqlite': ('org/xerial', 'sqlite-jdbc', '3.19.3'),
    }

    drivers = drivers or driverMap.keys()

    for d in drivers:
        try:
            src = driverMap[d]
        except KeyError:
            _log.errorExit('Unknown JDBC driver "{0}'.format(d))

        dst = ospath.join(lib_dir, '{0}.jar'.format(d))

        if ospath.exists(dst):
            continue

        if type(src) == tuple:
            src = '{0}/{1}/{2}/{3}/{2}-{3}.jar'.format(
                maven_central,
                src[0], src[1], src[2]
            )

        pathutil.downloadFile(env, src, dst)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext
from .. import log as _log


def _brew():
    return _ext.pathutil.which('brew')


def brewTap(tap):
    if not _ext.detect.isBrew():
        return

    brew = _brew()
    brew_sudo = _ext.os.environ.get('brewSudo', '').split()
    _ext.executil.callExternal(brew_sudo + [brew, 'tap', tap], cwd='/')


def brewUnlink(formula=None, search=None):
    if not _ext.detect.isBrew():
        return

    brew = _brew()
    brew_sudo = _ext.os.environ.get('brewSudo', '').split()

    flist = []

    if formula is not None:
        formula = _ext.configutil.listify(formula)
        flist += formula

    if search:
        flist += _ext.executil.callExternal([brew,
                                             'search', search], cwd='/').split()

    for f in flist:
        try:
            _ext.executil.callExternal(
                brew_sudo + [brew, 'unlink', f], cwd='/')
        except _ext.subprocess.CalledProcessError:
            _log.warn('You may need to unlink the formula manually!')


def brew(packages, cask=False):
    if not _ext.detect.isBrew():
        return

    packages = _ext.configutil.listify(packages)

    os = _ext.os
    brew = _brew()
    brew_sudo = os.environ.get('brewSudo', '').split()
    saved_mask = os.umask(0o022)

    try:
        for package in packages:
            try:
                if cask:
                    _ext.executil.callExternal(
                        brew_sudo + [brew, 'cask', 'install', package],
                        cwd='/',
                        user_interaction=True)
                elif brew == '/usr/local/bin/brew':
                    _ext.executil.callExternal(
                        brew_sudo + [brew, 'install',
                                     '--force-bottle', package],
                        cwd='/')
                else:
                    _ext.executil.callExternal(
                        brew_sudo + [brew, 'install', package],
                        cwd='/')
            except _ext.subprocess.CalledProcessError:
                _log.warn('You may need to enable the package manually')
    finally:
        os.umask(saved_mask)


def brewSearch(pattern, cask=False):
    if not _ext.detect.isBrew():
        return

    os = _ext.os
    brew = _brew()
    brew_sudo = os.environ.get('brewSudo', '').split()
    saved_mask = os.umask(0o022)

    pattern = '/{0}/'.format(pattern)

    if cask:
        cmd = [brew, 'cask', 'search', pattern]
    else:
        cmd = [brew, 'search', pattern]

    try:
        res = _ext.executil.callExternal(
            brew_sudo + cmd,
            cwd='/', verbose=False)
    finally:
        os.umask(saved_mask)

    return list(filter(None, res.strip().split('\n')))


def dmg(packages):
    if not _ext.detect.isMacOS():
        return

    packages = _ext.configutil.listify(packages)

    os = _ext.os
    path = _ext.pathutil
    glob = _ext.glob

    curl = _ext.pathutil.which('curl')
    hdiutil = _ext.pathutil.which('hdiutil')
    installer = _ext.pathutil.which('installer')
    volumes_dir = '/Volumes'

    for package in packages:
        base_name = package.split('/')[-1]
        local_name = os.path.join(os.environ['HOME'])

        # TODO: change to use env timeouts
        _ext.executil.callExternal([
            curl,
            '-fsSL',
            '--connect-timeout', '10',
            '--max-time', '300',
            '-o', base_name,
            package
        ])

        volumes = set(os.listdir(volumes_dir))
        _ext.executil.trySudoCall([hdiutil, 'attach', local_name])
        volume = (set(os.listdir(volumes_dir)) - volumes)[0]

        pkg = glob.glob(os.path.join(volumes_dir, volume, '*.pkg'))
        _ext.executil.trySudoCall([installer, '-package', pkg, '-target', '/'])

        _ext.executil.trySudoCall([hdiutil, 'dettach', local_name])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .data import DataSlots

from ..subtool import SubTool
from ..runtimetool import RuntimeTool
from ..rmstool import RmsTool
from ..vcstool import VcsTool


class ToolMixIn(DataSlots):
    __slots__ = ()

    def __init__(self):
        super(ToolMixIn, self).__init__()
        self._tool_impl = {}

    def _checkKnownTool(self, tool):
        if tool not in self._tool_impl:
            self._errorExit(
                'Implementation for "{0}" was not found'.format(tool))

    def _forEachTool(self, cb, allow_failure=False, base=None):
        config = self._config
        tools = config['toolOrder']

        for t in tools:
            t = self._getTool(t)

            if base and not isinstance(t, base):
                continue

            try:
                cb(config, t)
            except RuntimeError as e:
                if not allow_failure:
                    raise e

    def _getTool(self, name):
        self._checkKnownTool(name)
        tool_impl = self._tool_impl
        timpl = tool_impl[name]

        if isinstance(timpl, SubTool):
            return timpl

        tool_mod_name = timpl
        tool_module = self._ext.importlib.import_module(tool_mod_name)
        timpl = getattr(tool_module, name + 'Tool')(name)

        try:
            getattr(timpl, '__dict__')
            self._errorExit(
                'Missing __slots__ in {0}'.format(tool_mod_name))
        except KeyError:
            tool_impl[name] = timpl

        return timpl

    def _getKnownTools(self):
        return self._tool_impl.keys()

    def _getVcsTool(self):
        config = self._config
        vcs = config.get('vcs', None)

        if not vcs:
            self._errorExit(
                'Unknown VCS. Please set through --vcsRepo or project manifest')

        vcstool = self._getTool(vcs)

        if not config.get('vcsRepo', None):  # also check it set
            try:
                config['vcsRepo'] = vcstool.vcsGetRepo(config)
            except self._ext.subprocess.CalledProcessError as e:
                self._warn(str(e))

            if not config.get('vcsRepo', None):
                self._errorExit(
                    'Unknown VCS repo. Please set through --vcsRepo or project manifest')

        # Make sure these are present event after config is re-read
        self._overrides['vcs'] = vcs
        self._overrides['vcsRepo'] = config['vcsRepo']

        return vcstool

    def _getRmsTool(self):
        config = self._config
        rms = config.get('rms', None)

        if not rms:
            self._errorExit(
                'Unknown RMS. Please set through --rmsRepo or project manifest')

        if not config.get('rmsRepo', None):  # also check it set
            self._errorExit(
                'Unknown RMS repo. Please set through --rmsRepo or project manifest')

        # Make sure these are present event after config is re-read
        self._overrides['rms'] = rms
        self._overrides['rmsRepo'] = config['rmsRepo']

        return self._getTool(rms)

    def _getTarTool(self, compressor=None):
        env = self._env

        tar_tool = self._getTool('tar')
        tar_tool.requireInstalled(env)

        if compressor:
            self._getTool(compressor).requireInstalled(env)

        return tar_tool

    def _initTools(self):
        config = self._config
        env = self._env
        os = self._os
        ospath = self._ospath
        importlib = self._ext.importlib

        if config is None:
            config = self._overrides.copy()
            config['env'] = env

        #---
        config['projectRootSet'] = set(os.listdir('.'))

        #---
        plugins = {}
        plugins.update(env.get('plugins', {}))
        plugins.update(config.get('plugins', {}))

        plugin_packs = []
        plugin_packs += env.get('pluginPacks', [])
        plugin_packs += config.get('pluginPacks', [])
        plugin_packs.append('futoin.cid.tool')

        for pack_mod_name in plugin_packs:
            m = importlib.import_module(pack_mod_name)
            pack_dir = ospath.dirname(m.__file__)
            tool_files = os.listdir(pack_dir)
            tool_files = self._ext.fnmatch.filter(tool_files, '*tool.py')

            for f in tool_files:
                tool = f.replace('tool.py', '')
                tool_mod_name = '{0}.{1}tool'.format(pack_mod_name, tool)
                plugins[tool] = tool_mod_name

        tool_impl = self._tool_impl

        for (tool, t) in tool_impl.items():
            if isinstance(t, SubTool):
                t.onConfigReset()

        for (tool, tool_mod_name) in plugins.items():
            if tool not in tool_impl:
                tool_impl[tool] = tool_mod_name

        #---
        config_tools = config.get('tools', None)

        if config_tools:
            if not isinstance(config_tools, dict):
                self._errorExit(
                    'futoin.json:tools must be a map of tool=>version pairs')

            for (tool, v) in config_tools.items():
                self._checkKnownTool(tool)

                if v != '*' and v != True:
                    env[tool + 'Ver'] = v

        curr_tool = config.get('tool', None)

        if curr_tool:
            tools = [curr_tool]
            tool_ver = config.get('toolVer', None)

            if tool_ver:
                env[curr_tool + 'Ver'] = tool_ver
        else:
            config_tools = config.get('tools', None)
            tools = []

            if config_tools:
                tools.extend(config_tools.keys())
            elif self._project_config is not None and config.get('toolDetect', True):
                for n in self._getKnownTools():
                    t = self._getTool(n)
                    if t.autoDetect(config):
                        tools.append(n)

            # Make sure deps & env are processed for cli-supplied tools
            #--
            for (item, base) in {'rms': RmsTool, 'vcs': VcsTool}.items():
                tool = config.get(item, None)

                if tool:
                    tools.append(tool)

                    if not isinstance(self._getTool(tool), base):
                        self._errorExit(
                            'Tool {0} does not suite {1} type'.format(tool, item))

            # Make sure tools defined in entryPoints are auto-detected
            #--
            for (ep, ed) in config.get('entryPoints', {}).items():
                tool = ed.get('tool', None)

                if tool:
                    tools.append(tool)

                    if not isinstance(self._getTool(tool), RuntimeTool):
                        self._errorExit(
                            'Tool {0} does not suite RuntimeTool type'.format(tool))

        # add all deps
        #--
        dep_generations = [set(tools)]
        tools = set(tools)
        postdeps = set()
        dep_length = 0
        last_index = 0
        while len(dep_generations) != dep_length:
            dep_length = len(dep_generations)
            curr_index = last_index
            last_index = len(dep_generations)

            for g in dep_generations[curr_index:]:
                for tn in g:
                    t = self._getTool(tn)
                    moredeps = set(t.getDeps())
                    if moredeps:
                        dep_generations.append(moredeps)
                        tools.update(moredeps)
                    postdeps.update(set(t.getPostDeps()) - tools)

            if len(dep_generations) == dep_length and postdeps:
                dep_generations.append(postdeps)
                tools.update(postdeps)
                postdeps = set()

        #---
        if self._detect.isMacOS() and tools:
            # Make sure Homebrew is always implicit first tool
            dep_generations.append(set(['brew']))
            self._getTool('brew')

        #---
        dep_generations.reverse()
        tools = []
        for d in dep_generations:
            tools.extend(d - set(tools))
        config['toolOrder'] = tools

        #--
        for tool in tools:
            t = tool_impl[tool]
            t.envDeps(env)

        #--
        if config.get('toolTest', False):
            for tool in tools:
                t = tool_impl[tool]
                t.sanitizeVersion(env)
                t.importEnv(env)
                if not t.isInstalled(env):
                    break
        else:
            # note, it may have some undesired effect on parallel execution,
            # but let's leave that for now
            self._globalLock()

            for tool in tools:
                t = tool_impl[tool]
                t.sanitizeVersion(env)
                t.requireInstalled(env)
                if tool != curr_tool:
                    t.loadConfig(config)

            self._globalUnlock()

        # Solves generic issues of ordering independent tools in
        # later execution with predictable results:
        # 1. sort by integer order
        # 2. sort by tool name
        tools.sort(key=lambda v: (tool_impl[v].getOrder(), v))



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ..mixins.log import LogMixIn
from ..mixins.ondemand import OnDemandMixIn

from collections import OrderedDict


class ResourceAlgo(LogMixIn, OnDemandMixIn):
    CID_MIN_MEMORY = '48M'

    def pageSize(self):
        return self._os.sysconf('SC_PAGE_SIZE')

    def systemMemory(self):
        try:
            return self.pageSize() * self._os.sysconf('SC_PHYS_PAGES')
        except ValueError:
            if self._detect.isMacOS():
                return int(self._ext.subprocess.check_output(['sysctl', '-n', 'hw.memsize']).strip())
            else:
                self._errorExit('Failed to detect system memory size')

    def cgroupMemory(self, cgroupFile=None):
        cgroupFile = cgroupFile or '/sys/fs/cgroup/memory/memory.limit_in_bytes'

        if self._ospath.exists(cgroupFile):
            return int(self._pathutil.readTextFile(cgroupFile).strip())

        return None

    def memoryLimit(self, config):
        maxTotalMemory = config.get('deploy', {}).get('maxTotalMemory', None)

        if maxTotalMemory:
            return self._configutil.parseMemory(maxTotalMemory)

        sysMem = self.systemMemory()
        cgroupMem = self.cgroupMemory()

        if cgroupMem and cgroupMem < sysMem:
            return cgroupMem
        else:
            return sysMem / 2

    def systemCpuCount(self):
        if self._detect.isMacOS():
            return int(self._ext.subprocess.check_output(['sysctl', '-n', 'hw.ncpu']).strip())

        os = self._os
        return min(os.sysconf('SC_NPROCESSORS_ONLN'), os.sysconf('SC_NPROCESSORS_CONF'))

    def cgroupCpuCount(self, cgroupFile=None):
        cgroupFile = cgroupFile or '/sys/fs/cgroup/cpuset/cpuset.cpus'

        if not self._ospath.exists(cgroupFile):
            return None

        cpus = self._pathutil.readTextFile(cgroupFile).strip()
        cpus = cpus.split(',')
        count = 0

        for c in cpus:
            c = c.split('-')

            if len(c) == 2:
                count += len(range(int(c[0]), int(c[1]) + 1))
            else:
                count += 1

        return count

    def cpuLimit(self, config):
        maxCpuCount = config.get('deploy', {}).get('maxCpuCount', None)

        if maxCpuCount:
            return maxCpuCount

        cpu_count = self.systemCpuCount()
        cgroup_count = self.cgroupCpuCount()

        if cgroup_count:
            return min(cpu_count, cgroup_count)
        else:
            return cpu_count

    def configServices(self, config):
        memLimit = self.memoryLimit(config)
        cpuLimit = self.cpuLimit(config)

        self.distributeResources(config, memLimit, cpuLimit)
        self.assignSockets(config)

    def distributeResources(self, config, maxmem, maxcpu, granularity=None):
        import math
        configutil = self._configutil

        if granularity is None:
            granularity = self.pageSize()

        maxmem = maxmem // granularity
        availMem = maxmem
        deploy = config.setdefault('deploy', {})
        autoServices = deploy.setdefault('autoServices', {})
        entryPoints = config.get('entryPoints', {})

        services = {}
        candidates = set()
        env = config.get('env', {})
        debug = (env.get('type', 'dev') == 'dev')
        external_services = env.get('externalServices', [])
        cid_min_memory = configutil.parseMemory(
            self.CID_MIN_MEMORY) // granularity

        # Init
        for (en, ei) in entryPoints.items():
            tool = ei['tool']
            ei = ei.get('tune', {}).copy()

            for f in ('minMemory', 'connMemory', 'maxRequestSize'):
                if f not in ei:
                    self._errorExit(
                        '"{0}" is missing from {1} entry point'.format(f, en))
                ei[f] = configutil.parseMemory(ei[f]) // granularity

            for f in ('maxMemory', 'maxTotalMemory', 'debugOverhead', 'debugConnOverhead'):
                if f in ei:
                    ei[f] = configutil.parseMemory(ei[f]) // granularity

            for f in ('socketProtocol',):
                if f not in ei:
                    self._errorExit(
                        '"{0}" is missing from {1} entry point'.format(f, en))

            ei.setdefault('connFD', 16)
            ei.setdefault('memWeight', 100)
            ei.setdefault('cpuWeight', 100)
            ei.setdefault('maxMemory', maxmem)
            ei.setdefault('maxTotalMemory', maxmem)
            ei.setdefault('scalable', True)
            ei.setdefault('multiCore', True)
            ei.setdefault('reloadable', False)

            if debug:
                ei['minMemory'] += ei.get('debugOverhead', 0)
                ei['connMemory'] += ei.get('debugConnOverhead', 0)

            if (not ei['scalable']) and ei['maxMemory'] < ei['maxTotalMemory']:
                ei['maxTotalMemory'] = ei['maxMemory']

            ei['instances'] = 1
            ei['memAlloc'] = max(
                ei['minMemory'] + ei['connMemory'], cid_min_memory)

            disabled = tool in external_services
            ei['disabled'] = disabled

            if not disabled:
                availMem -= ei['memAlloc']
                candidates.add(en)

            services[en] = ei

        if availMem < 0:
            self._errorExit(
                'Not enough memory to allocate services: deficit "{0}" bytes'.format(availMem))

        # Distribute remaining
        while availMem > 0 and len(candidates) > 0:
            overall_weight = 0

            for en in candidates:
                overall_weight += services[en]['memWeight']

            distMem = availMem
            to_del = set()

            for en in candidates:
                ei = services[en]
                memAlloc = ei['memAlloc']
                addAlloc = distMem * ei['memWeight'] // overall_weight

                if (memAlloc + addAlloc) > ei['maxTotalMemory']:
                    ei['memAlloc'] = ei['maxTotalMemory']
                    addAlloc = ei['memAlloc'] - memAlloc
                    to_del.add(en)
                else:
                    ei['memAlloc'] += addAlloc

                availMem -= addAlloc

            candidates -= to_del

            if availMem == distMem:
                break

        # Distribute instances
        min_mem_coeff = 2

        for (en, ei) in services.items():
            if ei['disabled']:
                ei['maxCpuCount'] = '1'
                continue

            if not ei['scalable']:
                ei['maxCpuCount'] = maxcpu
                continue

            reasonableMinMemory = ei['minMemory'] * min_mem_coeff
            reasonableMinMemory = max(reasonableMinMemory, cid_min_memory)

            if ei['multiCore']:
                if (not ei['reloadable']) and ei['memAlloc'] >= (reasonableMinMemory * 2):
                    ei['instances'] = 2
                    ei['maxCpuCount'] = int(math.ceil(maxcpu / 2.0))
                else:
                    ei['maxCpuCount'] = maxcpu
            else:
                possible_instances = ei['memAlloc'] // reasonableMinMemory

                if possible_instances == 0 and ei['memAlloc'] > ei['minMemory']:
                    possible_instances = 1

                if ei['reloadable'] or maxcpu > 1:
                    ei['instances'] = min(maxcpu, possible_instances)
                else:
                    ei['instances'] = min(2, possible_instances)

                ei['maxCpuCount'] = 1

            ei['instances'] = min(ei['instances'], ei.get(
                'maxInstances', ei['instances']))

        # create services
        for (en, ei) in services.items():
            instances = []
            instance_count = ei['instances']
            service_mem = 0

            if instance_count <= 0:
                self._errorExit(
                    'Failed to allocate instances for "{0}"'.format(en))
            elif ei['disabled']:
                instance_count = 0

            for i in range(0, instance_count):
                ic = OrderedDict()
                instance_mem = ei['memAlloc'] // instance_count
                ic['maxMemory'] = instance_mem
                ic['connMemory'] = ei['connMemory']
                service_mem += instance_mem
                instances.append(ic)

            if instance_count > 0:
                instances[0]['maxMemory'] += (ei['memAlloc'] - service_mem)

            for ic in instances:
                ic['maxConnections'] = int(
                    ic['maxMemory'] - ei['minMemory']) // ei['connMemory']

                ic['maxFD'] = int(ei['connFD'] * ic['maxConnections'])
                ic['maxCpuCount'] = ei['maxCpuCount']
                ic['maxRequestSize'] = ei['maxRequestSize']
                ic['socketProtocol'] = ei['socketProtocol']

                for m in ('maxMemory', 'connMemory', 'maxRequestSize'):
                    ic[m] = configutil.toMemory(ic[m] * granularity)

            autoServices[en] = instances

    def assignSockets(self, config):
        port = 1025
        ports = set()
        deploy = config['deploy']
        autoServices = deploy['autoServices']
        entryPoints = config.get('entryPoints', {})

        ospath = self._ospath
        pathutil = self._pathutil
        os = self._os
        base_dir = ospath.realpath(config['deployDir'])
        run_dir = pathutil.safeJoin(base_dir, '.runtime')
        run_dir = deploy.setdefault('runtimeDir', run_dir)
        tmp_dir = pathutil.safeJoin(base_dir, '.tmp')
        tmp_dir = deploy.setdefault('tmpDir', tmp_dir)

        for (en, instances) in autoServices.items():
            ei = entryPoints[en].get('tune', {})

            if 'socketType' in ei and 'socketTypes' not in ei:
                self._errorExit(
                    'Entry point "{0}" sets socketType without socketTypes'.format(en))

            socket_types = ei.get('socketTypes', [])

            if not socket_types:
                continue

            sock_type = ei.get('socketType', socket_types[0])

            for i in range(0, len(instances)):
                ic = instances[i]

                ic['socketType'] = sock_type

                if sock_type == 'unix':
                    ic['socketPath'] = pathutil.safeJoin(
                        run_dir, '{0}.{1}.sock'.format(en, i))
                else:
                    ic['socketAddress'] = deploy.get(
                        'listenAddress', '0.0.0.0')

                    if 'socketPort' in ei:
                        sock_port = int(ei['socketPort']) + i
                    else:
                        sock_port = port
                        port += 1

                    if sock_port in ports:
                        self._errorExit(
                            'Port "{0}" is already in use'.format(sock_port))

                    ic['socketPort'] = sock_port
                    ports.add(sock_port)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool


class pythonTool(RuntimeTool):
    """Python is a programming language.

Home: https://www.python.org/

Only the first part of pythonVer is used for installation of
system packages OS-specific way.
"""
    __slots__ = ()

    VER_CMD = 'import sys; print( "%s.%s" % (sys.version_info.major, sys.version_info.minor) )'

    def getPostDeps(self):
        # hackish hack
        return ['virtualenv']

    def envNames(self):
        return ['pythonBin', 'pythonVer']

    def _installTool(self, env):
        if int(env['pythonVer'].split('.')[0]) == 3:
            self._install.deb(['python3'])
            self._install.zypper(['python3'])

            self._install.yumEPEL()
            self._install.yum(['python34'])
            self._install.pacman(['python'])
            self._install.apk('python3')
            self._install.brew('python3')
        else:
            self._install.debrpm(['python'])
            self._install.pacman(['python2'])
            self._install.apk('python2')
            self._install.brew('python')

        self._install.emerge(
            ['=dev-lang/python-{0}*'.format(env['pythonVer'])])

    def uninstallTool(self, env):
        pass

    def initEnv(self, env):
        environ = self._environ
        detect = self._detect

        python_ver = env.setdefault('pythonVer', '3')
        bin_name = None

        if detect.isGentoo():
            if len(python_ver) == 3:
                environ['EPYTHON'] = 'python{0}'.format(python_ver)
            elif int(python_ver.split('.')[0]) == 3:
                environ['EPYTHON'] = 'python3.4'
            else:
                environ['EPYTHON'] = 'python2.7'
        elif detect.isArchLinux():
            if int(python_ver.split('.')[0]) == 2:
                bin_name = 'python2'
        elif int(python_ver.split('.')[0]) == 3:
            bin_name = 'python3'

        super(pythonTool, self).initEnv(env, bin_name)

        if self._have_tool and 'pythonRawBin' not in env:
            env['pythonRawBin'] = env['pythonBin']
            python_ver_fact = self._executil.callExternal(
                [env['pythonRawBin'], '-c', self.VER_CMD],
                verbose=False
            ).strip()

            if python_ver.split('.') > python_ver_fact.split('.'):
                self._errorExit(
                    'Too old python version "{0}" when "{1}" is required'
                    .format(python_ver, python_ver_fact)
                )

            env['pythonFactVer'] = python_ver_fact

    def tuneDefaults(self, env):
        return {
            'minMemory': '8M',
            'socketType': 'none',
            'scalable': False,
            'reloadable': False,
            'multiCore': True,
        }



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class CurlToolMixIn(object):
    __slots__ = ()

    def getDeps(self):
        return ['curl']

    def _callCurl(self, env, curl_args, *args, **nargs):
        cmd = [env['curlBin'], '-fsSL']
        cmd += self._configutil.timeouts(env, 'curl')
        cmd += ['-A', 'FutoIn CID']
        cmd += curl_args

        return self._executil.callExternal(
            cmd,
            *args, **nargs
        )



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext


def apk(packages):
    if _ext.detect.isApk():
        packages = _ext.configutil.listify(packages)

        apk = '/sbin/apk'

        _ext.executil.trySudoCall(
            [apk, 'add'] + packages,
            errmsg='you may need to install the build deps manually !'
        )


def apkRepo(url, tag=None):
    if not _ext.detect.isApk():
        return

    apk = '/sbin/apk'
    repo_file = '/etc/apk/repositories'

    # version
    releasever = _ext.pathutil.readTextFile('/etc/alpine-release')
    releasever = releasever.strip().split('.')
    releasever = '.'.join(releasever[:2])
    repoline = url.replace('$releasever', releasever)

    # tag
    if tag:
        repoline = '@{0} {1}'.format(tag, repoline)

    repos = _ext.pathutil.readTextFile(repo_file).split("\n")
    repos = [r.strip() for r in repos]

    if repoline not in repos:
        _ext.executil.trySudoCall(
            ['/usr/bin/tee', '-a', repo_file],
            errmsg='you may need to add the repo manually!',
            input=repoline
        )
        _ext.executil.trySudoCall(
            [apk, 'update'],
            errmsg='you may need to update manually!'
        )


def apkCommunity():
    _ext.install.apkRepo(
        'http://dl-cdn.alpinelinux.org/alpine/v$releasever/community')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from ..testtool import TestTool
from .piptoolmixin import PipToolMixIn


class setuptoolsTool(PipToolMixIn, BuildTool, TestTool):
    """Easily download, build, install, upgrade, and uninstall Python packages.

Home: https://pypi.python.org/pypi/setuptools

Not assumed to be used directly.

Build targets:
    prepare -> {removes build & dist folders}
    build -> ['sdist', 'bdist_wheel']
    package -> {uses result of build in dist/}
Override targets with .config.toolTune.    

"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'setup.py'

    def uninstallTool(self, env):
        pass

    def initEnv(self, env):
        ospath = self._ospath
        virtualenv_dir = env['virtualenvDir']
        self._have_tool = ospath.exists(
            ospath.join(virtualenv_dir, 'bin', 'easy_install'))

    def onPrepare(self, config):
        targets = self._getTune(config, 'prepare', ['build', 'dist'])
        targets = self._configutil.listify(targets)

        for d in targets:
            self._pathutil.rmTree(d)

    def onBuild(self, config):
        env = config['env']
        self._requirePip(env, 'wheel')

        targets = self._getTune(config, 'build', ['sdist', 'bdist_wheel'])
        targets = self._configutil.listify(targets)

        cmd = [env['pythonBin'], 'setup.py'] + targets
        self._executil.callMeaningful(cmd)

    def onPackage(self, config):
        target = self._getTune(config, 'package', 'dist')
        self._pathutil.addPackageFiles(
            config, self._pathutil.safeJoin(target, '*'))

    def onCheck(self, config):
        env = config['env']
        self._requirePip(env, 'docutils')
        self._requirePip(env, 'readme')
        cmd = [env['pythonBin'], 'setup.py', 'check', '-mrs']
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import sys


class Coloring:
    __slots__ = ()
    _color = sys.stdout.isatty()

    @classmethod
    def enable(cls, val):
        cls._color = val

    @classmethod
    def wrap(cls, text, escape):
        if cls._color:
            ESC = chr(0x1B)
            return ESC + escape + str(text) + ESC + '[0m'
        else:
            return text

    @classmethod
    def error(cls, text):
        # bright red on black
        return cls.wrap(text, '[1;31m')

    @classmethod
    def warn(cls, text):
        # yellow on black
        return cls.wrap(text, '[33m')

    @classmethod
    def info(cls, text):
        # cyan on black
        return cls.wrap(text, '[36m')

    @classmethod
    def infoLabel(cls, text):
        # bright cyan on black
        return cls.wrap(text, '[1;36m')

    @classmethod
    def label(cls, text):
        # bright
        return cls.wrap(text, '[1m')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn


class virtualenvTool(BashToolMixIn, RunEnvTool):
    """Virtual Python Environment builder.

Home: https://pypi.python.org/pypi/virtualenv
"""
    __slots__ = ()

    def getDeps(self):
        return ['bash', 'python']

    def getPostDeps(self):
        # we need to ensure that if cid is called from virtualenv env
        # it's still can be used
        # So, CID can be in system and in each virtualenv as well
        return ['cid']

    def envNames(self):
        return ['virtualenvDir', 'virtualenvVer']

    def _installTool(self, env):
        python_ver = env['pythonVer'].split('.')

        virtualenv_dir = env['virtualenvDir']

        # virtualenv depends on ensurepip and provides no setuptools
        # ensurepip is not packaged on all OSes...
        if False:  # and int(python_ver[0]) == 3 and int(python_ver[1]) >= 3:
            self._executil.callExternal([
                env['pythonRawBin'],
                '-m', 'venv',
                '--system-site-packages',
                '--symlinks',
                '--without-pip',  # avoid ensurepip run
                '--clear',
                virtualenv_dir
            ])
        else:
            # Looks quite stupid, but it's a workaround for different OSes
            pip = self._pathutil.which('pip')

            if not pip:
                self._install.debrpm(['python-virtualenv'])
                self._install.debrpm(['virtualenv'])
                self._install.emerge(['dev-python/virtualenv'])
                self._install.pacman(['python-virtualenv'])
                self._install.apk(['py-virtualenv'])

                self._install.deb(['python-pip'])
                self._install.yum(['python2-pip'])
                self._install.emerge(['dev-python/pip'])
                self._install.pacman(['python-pip'])
                self._install.apk('py2-pip')

                if self._detect.isMacOS():
                    self._executil.trySudoCall(['easy_install', 'pip'])

                pip = self._pathutil.which('pip')

            if pip:
                pip_cmd = [pip, 'install', '-q', '--upgrade',
                           'virtualenv>={0}'.format(env['virtualenvVer'])]

                # TODO: use  --user without sudo
                self._executil.trySudoCall(pip_cmd)

            self._setupVirtualenv(env)

    def _setupVirtualenv(self, env):
        virtualenv = self._pathutil.which('virtualenv')

        if not virtualenv:
            self._errorExit('Failed to find virtualenv')

        self._executil.callExternal([
            virtualenv,
            '--python={0}'.format(env['pythonRawBin']),
            '--clear',
            env['virtualenvDir']
        ])

    def _afterExternalSetup(self, env):
        self._setupVirtualenv(env)

    def _updateTool(self, env):
        pass

    def initEnv(self, env):
        ospath = self._ospath
        virtualenv_dir = '.virtualenv-{0}'.format(env['pythonFactVer'])
        virtualenv_dir = env.setdefault(
            'virtualenvDir', ospath.join(self._pathutil.deployHome(), virtualenv_dir))

        env.setdefault('virtualenvVer', '15.1.0')

        self._have_tool = ospath.exists(
            ospath.join(virtualenv_dir, 'bin', 'activate'))

        if self._have_tool:
            env_to_set = self._callBash(env,
                                        'source {0}/bin/activate && env | grep \'{0}\''.format(
                                            virtualenv_dir),
                                        verbose=False
                                        )

            self._pathutil.updateEnvFromOutput(env_to_set)
            # reverse-dep hack
            env['pythonBin'] = ospath.join(virtualenv_dir, 'bin', 'python')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class brewTool(RunEnvTool):
    """Homebrew. The missing package manager for macOS.

Home: https://brew.sh/

brewInstall is use for admin user installation.
brewDir & brewGit is used for local install.

Hint: Unprivileged brew does not work well with many bottles, you may want to use
    brewSudo='/usr/bin/sudo -n -H -u adminaccount' with "cid sudoers" config.
"""
    __slots__ = ()

    _MACOS_ADMIN_GID = 80  # dirty hack for now
    _GLOBAL_BREW_DIR = '/usr/local'

    def envNames(self):
        return ['brewBin', 'brewDir', 'brewGit', 'brewInstall', 'brewSudo']

    def _installTool(self, env):
        if self._isLocalBrew(env):
            self._warn(
                'Unprivileged Homebrew install has many drawbacks. Check "cid tool describe brew"')
            homebrew_git = env['brewGit']
            homebrew_dir = env['brewDir']

            git = self._pathutil.which('git')

            if not git:
                xcode_select = self._pathutil.which('xcode-select')
                self._executil.callExternal(
                    [xcode_select, '--install'], suppress_fail=True)
                git = self._pathutil.which('git')

            self._executil.callExternal(
                [git, 'clone', homebrew_git, homebrew_dir])
        else:
            # should be system-available
            curl = self._pathutil.which('curl')
            ruby = self._pathutil.which('ruby')
            homebrew_install = env['brewInstall']

            curl_args = self._configutil.timeouts(env, 'curl')

            brew_installer = self._executil.callExternal(
                [curl, '-fsSL', homebrew_install] + curl_args
            )

            self._executil.callExternal([ruby, '-'], input=brew_installer)

    def _isLocalBrew(self, env):
        return env['brewDir'] != self._GLOBAL_BREW_DIR

        if self._MACOS_ADMIN_GID not in self._os.getgroups():
            return True

        return False

    def initEnv(self, env, bin_name=None):
        ospath = self._ospath
        os = self._os
        brewSudo = env.get('brewSudo', '')

        if self._MACOS_ADMIN_GID not in os.getgroups() and not brewSudo:
            homebrew_dir = ospath.join(self._environ['HOME'], '.homebrew')
            env.setdefault('brewDir', homebrew_dir)
        else:
            env.setdefault('brewDir', self._GLOBAL_BREW_DIR)

            if brewSudo:
                self._environ['brewSudo'] = brewSudo

        self._environ['HOMEBREW_NO_GITHUB_API'] = '1'
        env.setdefault('brewGit',
                       'https://github.com/Homebrew/brew.git')
        env.setdefault('brewInstall',
                       'https://raw.githubusercontent.com/Homebrew/install/master/install')

        if self._isLocalBrew(env):
            homebrew_dir = env['brewDir']
            bin_dir = ospath.join(homebrew_dir, 'bin')
            brew = ospath.join(bin_dir, 'brew')

            if ospath.exists(brew):
                self._pathutil.addBinPath(bin_dir, True)
                env['brewBin'] = brew
                self._have_tool = True
        else:
            env.setdefault('brewDir', self._GLOBAL_BREW_DIR)
            super(brewTool, self).initEnv(env, bin_name)



#!/usr/bin/env python

#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""FutoIn Continuous Integration & Delivery Tool.

Usage:
    cid init [<project_name>] [--vcsRepo=<vcs_repo>] [--rmsRepo=<rms_repo>] [--permissive]
    cid tag <branch> [<next_version>] [--vcsRepo=<vcs_repo>] [--wcDir=<wc_dir>]
    cid prepare [<vcs_ref>] [--vcsRepo=<vcs_repo>] [--wcDir=<wc_dir>]
    cid build [--debug]
    cid package
    cid check [--permissive]
    cid promote <rms_pool> <packages>... [--rmsRepo=<rms_repo>]
    cid deploy setup [--deployDir=<deploy_dir>] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>] [--runtimeDir=<runtime_dir>] [--tmpDir=<tmp_dir>] [--user=<user>] [--group=<group>]
    cid deploy vcstag [<vcs_ref>] [--vcsRepo=<vcs_repo>] [--redeploy] [--deployDir=<deploy_dir>] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>] [--runtimeDir=<runtime_dir>] [--tmpDir=<tmp_dir>] [--user=<user>] [--group=<group>]
    cid deploy vcsref <vcs_ref> [--vcsRepo=<vcs_repo>] [--redeploy] [--deployDir=<deploy_dir>] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>] [--runtimeDir=<runtime_dir>] [--tmpDir=<tmp_dir>] [--user=<user>] [--group=<group>]
    cid deploy rms <rms_pool> [<package>] [--rmsRepo=<rms_repo>] [--redeploy] [--deployDir=<deploy_dir>] [--build] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>] [--runtimeDir=<runtime_dir>] [--tmpDir=<tmp_dir>] [--user=<user>] [--group=<group>]
    cid deploy reset [<set_type>] [--deployDir=<deploy_dir>]
    cid deploy set tools <tools>... [--deployDir=<deploy_dir>]
    cid deploy set tooltune <tool> <tune>... [--deployDir=<deploy_dir>]
    cid deploy set action <name> <action>... [--deployDir=<deploy_dir>]
    cid deploy set persistent <path>... [--deployDir=<deploy_dir>]
    cid deploy set writable <path>... [--deployDir=<deploy_dir>]
    cid deploy set entrypoint <name> <tool> <entry_path> [<tune>...] [--deployDir=<deploy_dir>]
    cid deploy set env <variable> [<value>] [--deployDir=<deploy_dir>]
    cid deploy set webcfg <variable> [<value>] [--deployDir=<deploy_dir>]
    cid deploy set webmount <web_path> [<json>] [--deployDir=<deploy_dir>]
    cid migrate
    cid run
    cid run <command> [--] [<command_arg>...]
    cid ci_build <vcs_ref> [<rms_pool>] [--vcsRepo=<vcs_repo>] [--rmsRepo=<rms_repo>] [--permissive] [--debug] [--wcDir=<wc_dir>]
    cid tool exec <tool_name> [<tool_version>] [-- <tool_arg>...]
    cid tool envexec <tool_name> [<tool_version>] [-- <any_command>...]
    cid tool (install|uninstall|update|test|env) [<tool_name> [<tool_version>]]
    cid tool (prepare|build|check|package|migrate) <tool_name> [<tool_version>]
    cid tool list
    cid tool describe <tool_name>
    cid tool detect
    cid vcs checkout [<vcs_ref>] [--vcsRepo=<vcs_repo>] [--wcDir=<wc_dir>]
    cid vcs commit <commit_msg> [<commit_files>...] [--wcDir=<wc_dir>]
    cid vcs merge <vcs_ref> [--no-cleanup] [--wcDir=<wc_dir>]
    cid vcs branch <vcs_ref> [--wcDir=<wc_dir>]
    cid vcs delete <vcs_ref> [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs export <vcs_ref> <dst_dir> [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs tags [<tag_pattern>] [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs branches [<branch_pattern>] [--vcsRepo=<vcs_repo>] [--cacheDir=<cache_dir>] [--wcDir=<wc_dir>]
    cid vcs reset [--wcDir=<wc_dir>]
    cid vcs ismerged <vcs_ref> [--wcDir=<wc_dir>]
    cid vcs clean [--wcDir=<wc_dir>]
    cid rms list <rms_pool> [<package_pattern>] [--rmsRepo=<rms_repo>]
    cid rms retrieve <rms_pool> <packages>... [--rmsRepo=<rms_repo>]
    cid rms pool create <rms_pool> [--rmsRepo=<rms_repo>]
    cid rms pool list [--rmsRepo=<rms_repo>]
    cid devserve [--wcDir=<wc_dir>] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>]
    cid service master [--deployDir=<deploy_dir>] [--adapt] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>] [--user=<user>] [--group=<group>]
    cid service list [--deployDir=<deploy_dir>] [--adapt] [--limit-memory=<mem_limit>] [--limit-cpus=<cpu_count>] [--listen-addr=<address>] [--user=<user>] [--group=<group>]
    cid service exec <entry_point> <instance_id> [--deployDir=<deploy_dir>]
    cid service stop <entry_point> <instance_id> <pid> [--deployDir=<deploy_dir>]
    cid service reload <entry_point> <instance_id> <pid> [--deployDir=<deploy_dir>]
    cid sudoers [<sudo_entity>] [--skip-key-management]
    cid build-dep [<build_dep>...]
    

Options:
    -h --help                       Show this screen.
    --vcsRepo=<vcs_repo>            VCS repository URL in vcs_type:vcs_url format.
    --rmsRepo=<rms_repo>            RMS repository URL in rms_type:rms_url format.
    --wcDir=<wc_dir>                Working copy directory (project root).
    --rmsHash=<rms_hash>            Package hash for validation in hash_type:value format.
    --deployDir=<deploy_dir>        Destination for deployment.
    --redeploy                      Force redeploy.
    --build                         Build during deploy.
    --permissive                    Ignore test failures.
    --debug                         Build in debug mode, if applicable.
    --cacheDir=<cache_dir>          Directory to hold VCS cache.
    --adapt                         Re-balance available resources before execution.
    --limit-memory=<mem_limit>|auto Limit allocated memory (B, K, M or G postfix is required).
    --limit-cpus=<cpu_count>|auto   Limit CPU cores (affects instance count).
    --listen-addr=<address>|auto    Address to bind services, all by default.
    --runtimeDir=<runtime_dir>|auto Directory for runtime data (sockets, configs, etc.).
    --tmpDir=<tmp_dir>|auto         Directory for temporary data.
    --user=<user>|auto              User name to use for service execution.
    --group=<group>|auto            Group name to use for service execution.
    --skip-key-management           Enforce better security forbidding signing key management.
    <rms_pool>                      Either "dst_pool" or  "src_pool:dst_pool" for promotion.
    <packages>                      Either "local/file" or "remote_file" or "<package>@<hash>".
    <sudo_entity>                   Username or group to be put in generated sudoers as is.
"""

from __future__ import print_function

from .cidtool import CIDTool
from .coloring import Coloring

try:
    from docopt import docopt
except ImportError:
    # fallback to "hardcoded"
    from futoin.cid.contrib.docopt import docopt

import os
import sys

if sys.version_info < (2, 7):
    print('Sorry, but only Python version >= 2.7 is supported!', file=sys.stderr)
    sys.exit(1)

from . import __version__ as version
__all__ = ['run']


def runInner():
    args = docopt(__doc__, version='FutoIn CID v{0}'.format(version))

    if type(args) == str:
        print(args)
    else:
        ospath = os.path

        if 'CID_COLOR' in os.environ:
            Coloring.enable(os.environ['CID_COLOR'] == 'yes')

        overrides = {}
        #---
        vcsArg = args.get('--vcsRepo', None)

        if vcsArg:
            try:
                (overrides['vcs'],
                 overrides['vcsRepo']) = vcsArg.split(':', 1)
            except ValueError:
                raise RuntimeError('Invalid argument to --vcsRepo')
        #---
        rmsArg = args.get('--rmsRepo', None)

        if rmsArg:
            try:
                (overrides['rms'],
                 overrides['rmsRepo']) = rmsArg.split(':', 1)
            except ValueError:
                raise RuntimeError('Invalid argument to --rmsRepo')
        overrides['rmsPool'] = args['<rms_pool>']

        #---
        if args['ci_build']:
            if 'vcs' in overrides:
                def_wc_dir = 'ci_build'
            else:
                def_wc_dir = ospath.join('..', 'ci_builds',
                                         ospath.basename(ospath.realpath('.')))
            def_wc_dir += '_' + args['<vcs_ref>']
        else:
            def_wc_dir = '.'
        overrides['wcDir'] = ospath.realpath(args['--wcDir'] or def_wc_dir)
        #--
        deploy_dir = args['--deployDir']

        if deploy_dir:
            overrides['deployDir'] = ospath.realpath(deploy_dir)
        elif args['deploy'] or args['service']:
            overrides['deployDir'] = ospath.realpath('.')

        overrides['reDeploy'] = args['--redeploy'] and True or False

        if args['--build']:
            overrides['deployBuild'] = True

        if args['--debug']:
            overrides['debugBuild'] = True

        # enable vcsref & vcstag build by default
        if args['deploy'] and (args['vcsref'] or args['vcstag']):
            overrides['deployBuild'] = True

        #---
        overrides['vcsRef'] = args['<vcs_ref>']

        #---
        tool = args['<tool_name>']
        tool_ver = args['<tool_version>']
        overrides['tool'] = tool
        overrides['toolVer'] = tool_ver != '--' and tool_ver or None
        overrides['toolTest'] = args['tool'] and (
            args['test'] or
            args['uninstall'] or
            args['describe'] or
            args['detect'] or
            args['env'] or
            (args['deploy'] and args['set'])
        )

        overrides['toolDetect'] = not (args['service'] and args['master'])

        #---
        if args['--permissive']:
            overrides['permissiveChecks'] = True

        #---
        cache_dir = args['--cacheDir']

        if cache_dir:
            cache_dir = ospath.realpath(cache_dir)

        #---
        deploy = overrides.setdefault('_deploy', {})
        deploy['maxTotalMemory'] = args['--limit-memory']
        deploy['maxCpuCount'] = args['--limit-cpus']
        deploy['listenAddress'] = args['--listen-addr']
        deploy['runtimeDir'] = args['--runtimeDir']
        deploy['tmpDir'] = args['--tmpDir']
        deploy['user'] = args['--user']
        deploy['group'] = args['--group']
        overrides['adaptDeploy'] = args['--adapt'] or args['devserve']

        if deploy['maxCpuCount']:
            try:
                deploy['maxCpuCount'] = int(deploy['maxCpuCount'])
            except ValueError:
                pass

        #---
        cit = CIDTool(overrides=overrides)

        if args['tool']:
            if args['exec']:
                cit.tool_exec(tool, args['<tool_arg>'])
            if args['envexec']:
                cit.tool_envexec(tool, args['<any_command>'])
            elif args['list']:
                cit.tool_list()
            elif args['detect']:
                cit.tool_detect()
            else:
                subcmds = [
                    'install',
                    'uninstall',
                    'update',
                    'test',
                    'env',
                    'prepare',
                    'build',
                    'check',
                    'package',
                    'migrate',
                    'describe',
                ]
                for cmd in subcmds:
                    if args[cmd]:
                        getattr(cit, 'tool_' + cmd)(tool)
                        break
                else:
                    raise RuntimeError("Unknown Command")
        elif args['vcs']:
            if args['checkout']:
                cit.vcs_checkout(args['<vcs_ref>'])
            elif args['commit']:
                cit.vcs_commit(args['<commit_msg>'], args['<commit_files>'])
            elif args['branch']:
                cit.vcs_branch(args['<vcs_ref>'])
            elif args['merge']:
                cit.vcs_merge(args['<vcs_ref>'], not args['--no-cleanup'])
            elif args['delete']:
                cit.vcs_delete(args['<vcs_ref>'], cache_dir)
            elif args['export']:
                dst_dir = ospath.realpath(args['<dst_dir>'])
                cit.vcs_export(args['<vcs_ref>'], dst_dir, cache_dir)
            elif args['tags']:
                cit.vcs_tags(args['<tag_pattern>'], cache_dir)
            elif args['branches']:
                cit.vcs_branches(args['<branch_pattern>'], cache_dir)
            elif args['reset']:
                cit.vcs_reset()
            elif args['ismerged']:
                cit.vcs_ismerged(args['<vcs_ref>'])
            elif args['clean']:
                cit.vcs_clean()
            else:
                raise RuntimeError("Not implemented yet.")
        elif args['rms'] and not args['deploy']:
            if args['pool']:
                if args['create']:
                    cit.rms_pool_create(args['<rms_pool>'])
                elif args['list']:
                    cit.rms_pool_list()
                else:
                    raise RuntimeError("Not implemented yet.")
            elif args['list']:
                cit.rms_list(args['<rms_pool>'], args['<package_pattern>'])
            elif args['retrieve']:
                cit.rms_retrieve(args['<rms_pool>'], args['<packages>'])
            else:
                raise RuntimeError("Not implemented yet.")
        elif args['deploy']:
            if args['rms']:
                cit.deploy('rms', args['<rms_pool>'], args['<package>'])
            elif args['vcsref']:
                cit.deploy('vcsref', args['<vcs_ref>'])
            elif args['vcstag']:
                cit.deploy('vcstag', args['<vcs_ref>'])
            elif args['setup']:
                cit.deploy('setup')
            elif args['set']:
                if args['action']:
                    cit.deploy_set('action', args['<name>'], args['<action>'])
                elif args['persistent']:
                    cit.deploy_set('persistent', args['<path>'])
                elif args['writable']:
                    cit.deploy_set('writable', args['<path>'])
                elif args['entrypoint']:
                    cit.deploy_set('entrypoint', args['<name>'], args['<tool>'],
                                   args['<entry_path>'], args['<tune>'])
                elif args['env']:
                    cit.deploy_set('env', args['<variable>'], args['<value>'])
                elif args['webcfg']:
                    cit.deploy_set(
                        'webcfg', args['<variable>'], args['<value>'])
                elif args['webmount']:
                    cit.deploy_set(
                        'webmount', args['<web_path>'], args['<json>'])
                elif args['tools']:
                    cit.deploy_set('tools', args['<tools>'])
                elif args['tooltune']:
                    cit.deploy_set('tooltune', args['<tool>'], args['<tune>'])
                else:
                    raise RuntimeError("Not implemented yet.")
            elif args['reset']:
                cit.deploy_reset(args['<set_type>'])
            else:
                raise RuntimeError("Not implemented yet.")
        elif args['service']:
            if args['master']:
                cit.service_master()
            elif args['list']:
                cit.service_list()
            elif args['exec']:
                cit.service_exec(args['<entry_point>'], args['<instance_id>'])
            elif args['stop']:
                cit.service_stop(args['<entry_point>'],
                                 args['<instance_id>'], args['<pid>'])
            elif args['reload']:
                cit.service_reload(args['<entry_point>'],
                                   args['<instance_id>'], args['<pid>'])
            else:
                raise RuntimeError("Not implemented yet.")
        elif args['init']:
            cit.init_project(args['<project_name>'])
        elif args['tag']:
            cit.tag(args['<branch>'], args['<next_version>'])
        elif args['prepare']:
            cit.prepare(args['<vcs_ref>'])
        elif args['build']:
            cit.build()
        elif args['package']:
            cit.package()
        elif args['check']:
            cit.check()
        elif args['promote']:
            cit.promote(args['<rms_pool>'], args['<packages>'])
        elif args['migrate']:
            cit.migrate()
        elif args['run']:
            cit.run(args['<command>'], args['<command_arg>'])
        elif args['ci_build']:
            cit.ci_build(args['<vcs_ref>'], args['<rms_pool>'])
        elif args['devserve']:
            cit.devserve()
        elif args['sudoers']:
            cit.sudoers(args['<sudo_entity>'], args['--skip-key-management'])
        elif args['build-dep']:
            cit.builddep(args['<build_dep>'])
        else:
            raise RuntimeError("Unknown Command")


def run():
    try:
        runInner()
    except Exception as e:
        print(file=sys.stderr)
        print(Coloring.error('ERROR: ' + str(e)), file=sys.stderr)
        print(file=sys.stderr)
        import traceback
        print(Coloring.warn(traceback.format_exc()), file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt as e:
        print(file=sys.stderr)
        print(Coloring.error('Exit on user abort'), file=sys.stderr)
        print(file=sys.stderr)
        sys.exit(1)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class rustTool(RunEnvTool):
    """Rust is a systems programming language.

Home: https://www.rust-lang.org
"""
    __slots__ = ()

    def getDeps(self):
        if self._isGlobalRust():
            return []

        return ['rustup']

    def getVersionParts(self):
        return 3

    def _isGlobalRust(self):
        return self._detect.isAlpineLinux()

    def _installTool(self, env):
        if self._isGlobalRust():
            self._install.apk('rust')
            return

        self._executil.callExternal([
            env['rustupBin'], 'toolchain', 'install', env['rustVer']
        ])

    def _updateTool(self, env):
        self._installTool(env)

    def uninstallTool(self, env):
        if self._isGlobalRust():
            return

        self._executil.callExternal([
            env['rustupBin'], 'toolchain', 'uninstall', env['rustVer']
        ])
        self._have_tool = False

    def envNames(self):
        return ['rustBin', 'rustVer']

    def initEnv(self, env):
        if not self._isGlobalRust():
            ver = env.setdefault('rustVer', 'stable')
            self._environ['RUSTUP_TOOLCHAIN'] = ver

            try:
                res = self._executil.callExternal([
                    env['rustupBin'], 'which', 'rustc'
                ], verbose=False)
            except:
                return

        super(rustTool, self).initEnv(env, 'rustc')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext
from . import log as _log

_api_url = 'https://api.github.com'
_api_ver = 'application/vnd.github.v3+json'


def api_call(env, method, path, **kwargs):
    url = _api_url + path
    _ext.configutil.requestsOptions(env, kwargs)

    headers = kwargs.setdefault('headers', {})
    headers.setdefault('Accept', _api_ver)

    return _ext.requests.request(method, url, **kwargs)


def listReleases(env, repo):
    res = api_call(env, 'GET',
                   '/repos/{0}/releases'.format(repo))
    res.raise_for_status()
    return res.json()


def releaseInfo(env, repo, ver):
    res = api_call(env, 'GET',
                   '/repos/{0}/releases/{1}'.format(repo, ver))
    res.raise_for_status()
    return res.json()


def findAsset(assets, content_type):
    for asset in assets:
        if asset['content_type'] == content_type:
            return asset
    else:
        _log.errorExit('Failed to find .tar.gz asset')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..vcstool import VcsTool
from .bashtoolmixin import BashToolMixIn


class gitTool(BashToolMixIn, VcsTool):
    """Git distributed version control system.

Home: https://git-scm.com/

Git tool forcibly sets user.email and user.name, 
if not set by user.

Environment variables affecting behavior:
* gitSignTags=1 - non-empty value triggers tag signing instead of
                  simple annotation.
* gitIdentity=.. - non-empty value forced specific identify to be
                  used for tag signing.
"""
    __slots__ = ()

    def getDeps(self):
        return ['bash', 'tar']

    def autoDetectFiles(self):
        return '.git'

    def _installTool(self, env):
        self._install.debrpm(['git'])
        self._install.emerge(['dev-vcs/git'])
        self._install.pacman(['git'])
        self._install.apk(['git'])
        self._install.brew('git')

    def _checkGitConfig(self, env):
        gitBin = env['gitBin']
        user_email = None
        user_name = None

        try:
            user_email = self._executil.callExternal([
                gitBin, 'config', 'user.email',
            ], verbose=False).strip()
        except:
            pass

        try:
            user_name = self._executil.callExternal([
                gitBin, 'config', 'user.name',
            ], verbose=False).strip()
        except:
            pass

        if not user_email:
            self._info('Setting fallback user.email for repo')
            self._executil.callExternal([
                gitBin, 'config', 'user.email',
                env.get('gitUserEmail', 'noreply@futoin.org')
            ])

        if not user_name:
            self._info('Setting fallback user.name for repo')
            self._executil.callExternal([
                gitBin, 'config', 'user.name',
                env.get('gitUserName', 'FutoIn CITool')
            ])

    def _getCurrentBranch(self, config):
        return self._executil.callExternal([
            config['env']['gitBin'], 'rev-parse', '--abbrev-ref', 'HEAD'
        ], verbose=False).strip()

    def vcsGetRepo(self, config, wc_dir=None):
        git_dir = wc_dir or self._pathutil.safeJoin(self._os.getcwd(), '.git')

        return self._executil.callExternal([
            config['env']['gitBin'],
            '--git-dir={0}'.format(git_dir),
            'config',
            '--get',
            'remote.origin.url'
        ], verbose=False).strip()

    def _gitCompareRepo(self, cfg, act):
        return cfg == act or ('ssh://' + cfg) == act

    def vcsCheckout(self, config, vcs_ref):
        gitBin = config['env']['gitBin']
        wc_dir = self._os.getcwd()
        vcsRepo = config['vcsRepo']
        vcs_ref = vcs_ref or 'master'

        if self._ospath.exists('.git'):
            remote_url = self.vcsGetRepo(config, '.git')

            if not self._gitCompareRepo(vcsRepo, remote_url):
                self._errorExit("Git remote mismatch: '{0}' != '{1}'"
                                .format(vcsRepo, remote_url))

            self._executil.callExternal([gitBin, 'fetch', '-q'])
        else:
            self._executil.callExternal(
                [gitBin, 'clone', '-q', vcsRepo, wc_dir])

            try:
                self._executil.callExternal(
                    [gitBin, 'rev-parse', 'HEAD'], verbose=False)
            except self._ext.subprocess.CalledProcessError:
                # exit on empty repository
                return

        remote_branch = self._executil.callExternal([
            gitBin, 'branch', '-q', '--all', '--list', 'origin/' + vcs_ref
        ], verbose=False).strip()

        if self._executil.callExternal([gitBin, 'branch', '-q', '--list', vcs_ref], verbose=False).strip():
            self._executil.callExternal([gitBin, 'checkout', '-q', vcs_ref])

            if remote_branch:
                self._executil.callExternal(
                    [gitBin, 'branch', '-q', '--set-upstream-to', 'origin/' + vcs_ref])
                self._executil.callExternal(
                    [gitBin, 'rebase', 'origin/' + vcs_ref])
        elif remote_branch:
            self._executil.callExternal(
                [gitBin, 'checkout', '-q', '--track', '-b', vcs_ref, 'origin/' + vcs_ref])
        else:
            self._executil.callExternal([gitBin, 'checkout', '-q', vcs_ref])

    def vcsCommit(self, config, message, files):
        env = config['env']
        gitBin = env['gitBin']
        self._checkGitConfig(env)

        if files:
            self._executil.callExternal([gitBin, 'add'] + files)
        else:
            self._executil.callExternal([gitBin, 'add', '-A'])
            files = []

        self._executil.callExternal(
            [gitBin, 'commit', '-q', '-m', message] + files)

    def vcsTag(self, config, tag, message):
        env = config['env']

        gitSignTags = env.get('gitSignTags', '')
        gitIdentity = env.get('gitIdentity', '')

        if gitIdentity:
            sign_arg = '--local-user={0}'.format(gitIdentity)
        elif gitSignTags:
            sign_arg = '-s'
        else:
            sign_arg = '-a'

        gitBin = env['gitBin']
        self._executil.callExternal(
            [gitBin, 'tag', sign_arg, '-m', message, tag])

    def vcsPush(self, config, refs, repo=None, check_empty=True):
        refs = refs or []
        gitBin = config['env']['gitBin']
        repo = repo or 'origin'

        if check_empty:
            try:
                self._executil.callExternal(
                    [gitBin, 'rev-parse', 'HEAD'], verbose=False)
            except self._ext.subprocess.CalledProcessError:
                # exit on empty repository
                return

        self._executil.callExternal(
            [gitBin, '-c', 'push.default=current', 'push', '-q', repo] + refs)

    def vcsGetRevision(self, config):
        gitBin = config['env']['gitBin']
        return self._executil.callExternal([
            gitBin, 'rev-parse', 'HEAD'
        ], verbose=False).strip()

    def vcsGetRefRevision(self, config, vcs_cache_dir, branch):
        res = self._executil.callExternal([
            config['env']['gitBin'],
            'ls-remote', '--refs',
            config['vcsRepo'],
            'refs/heads/{0}'.format(branch)
        ], verbose=False).strip()

        if res:
            return res.split()[0]

        self._errorExit("Uknown Git ref: '{0}'".format(branch))

    def vcsListTags(self, config, vcs_cache_dir, tag_hint):
        if tag_hint:
            tag_hint = ['refs/tags/{0}'.format(tag_hint)]
        else:
            tag_hint = []

        res = self._executil.callExternal([
            config['env']['gitBin'],
            'ls-remote', '--tags', '--refs',
            config['vcsRepo']
        ] + tag_hint, verbose=False).strip().split("\n")

        res = [v and v.split()[1].replace('refs/tags/', '') or '' for v in res]
        res = list(filter(None, res))
        return res

    def vcsListBranches(self, config, vcs_cache_dir, branch_hint):
        if branch_hint:
            branch_hint = ['refs/heads/{0}'.format(branch_hint)]
        else:
            branch_hint = []

        res = self._executil.callExternal([
            config['env']['gitBin'],
            'ls-remote', '--heads', '--refs',
            config['vcsRepo']
        ] + branch_hint, verbose=False).strip().split("\n")

        res = [v and v.split()[1].replace(
            'refs/heads/', '') or '' for v in res]
        res = list(filter(None, res))
        return res

    def vcsExport(self, config, vcs_cache_dir, vcs_ref, dst_path):
        ospath = self._ospath
        env = config['env']
        gitBin = env['gitBin']
        vcsRepo = config['vcsRepo']

        if vcs_cache_dir is None:
            cache_repo = vcsRepo
        else:
            if ospath.exists(vcs_cache_dir):
                remote_url = self.vcsGetRepo(config, vcs_cache_dir)

                if not self._gitCompareRepo(vcsRepo,  remote_url):
                    self._warn('removing git cache on remote URL mismatch: {0} != {1}'
                               .format(remote_url, vcsRepo))
                    self._pathutil.rmTree(vcs_cache_dir)

            if not ospath.exists(vcs_cache_dir):
                self._executil.callExternal([
                    env['gitBin'],
                    'clone',
                    '--mirror',
                    '--depth=1',
                    '--no-single-branch',
                    vcsRepo,
                    vcs_cache_dir
                ])
            else:
                self._executil.callExternal([
                    env['gitBin'],
                    '--git-dir={0}'.format(vcs_cache_dir),
                    'fetch'
                ])

            cache_repo = 'file://' + vcs_cache_dir

        self._pathutil.rmTree(dst_path)
        self._os.makedirs(dst_path)

        self._callBash(env, '{0} archive --remote={1} --format=tar {2} | {3} x -C {4}'
                       .format(config['env']['gitBin'], cache_repo, vcs_ref, env['tarBin'], dst_path))

    def vcsBranch(self, config, vcs_ref):
        env = config['env']
        gitBin = env['gitBin']

        self._executil.callExternal([
            config['env']['gitBin'],
            'checkout', '-b', vcs_ref,
        ])

        self.vcsPush(config, [vcs_ref])

    def vcsMerge(self, config, vcs_ref, cleanup):
        curr_ref = self._getCurrentBranch(config)

        env = config['env']
        gitBin = env['gitBin']

        try:
            self._executil.callExternal([
                config['env']['gitBin'],
                'merge', '--no-ff', 'origin/' + vcs_ref,
            ])
        except self._ext.subprocess.CalledProcessError:
            if cleanup:
                self.vcsRevert(config)
                self._errorExit('Merge failed, aborted.')
            self._errorExit('Merge failed, left as-is.')

        self.vcsPush(config, [curr_ref])

    def vcsDelete(self, config, vcs_cache_dir, vcs_ref):
        env = config['env']
        gitBin = env['gitBin']

        have_local = (self._ospath.exists('.git') and
                      self._gitCompareRepo(config['vcsRepo'],  self.vcsGetRepo(config)))

        if have_local:
            try:
                self._executil.callExternal([
                    config['env']['gitBin'],
                    'branch', '-D', vcs_ref,
                ])
            except self._ext.subprocess.CalledProcessError as e:
                self._warn(str(e))

            self.vcsPush(config, ['--force', '--delete', vcs_ref])

            self._executil.callExternal([
                config['env']['gitBin'],
                'fetch', '--all', '--prune',
            ])
        else:
            # a dirty hack to avoid error message of not found
            # local git repo
            repo = self._pathutil.tmpCacheDir(prefix='git_')

            self._executil.callExternal([
                config['env']['gitBin'],
                'init', repo,
            ], verbose=False)

            os = self._os
            oldcwd = os.getcwd()
            os.chdir(repo)
            self.vcsPush(config, ['--force', '--delete',
                                  vcs_ref], config['vcsRepo'], False)
            os.chdir(oldcwd)

            self._pathutil.rmTree(repo)

    def vcsRevert(self, config):
        ospath = self._ospath

        if ospath.exists(ospath.join('.git', 'MERGE_HEAD')):
            self._executil.callExternal([
                config['env']['gitBin'],
                'merge', '--abort',
            ])

        self._executil.callExternal([
            config['env']['gitBin'],
            'reset', '--hard',
        ])

    def vcsIsMerged(self, config, vcs_ref):
        res = self._executil.callExternal([
            config['env']['gitBin'],
            'branch', '-r', '--merged', 'HEAD', 'origin/{0}'.format(vcs_ref)
        ], verbose=False).strip()
        return res != ''

    def vcsClean(self, config):
        self._executil.callExternal([
            config['env']['gitBin'],
            'clean', '-f', '-x', '-d',
        ])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .subtool import SubTool

__all__ = ['RmsTool']


class RmsTool(SubTool):
    __slots__ = ()

    ALLOWED_HASH_TYPES = [
        'md5',
        'sha1',
        'sha256',
        'sha512',
    ]

    def autoDetect(self, config):
        if self._autoDetectRMS(config):
            return True

        return super(RmsTool, self).autoDetect(config)

    def rmsUpload(self, config, rms_pool, package_list):
        raise NotImplementedError(self._name)

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        raise NotImplementedError(self._name)

    def rmsGetList(self, config, rms_pool, package_hint):
        raise NotImplementedError(self._name)

    def rmsRetrieve(self, config, rms_pool, package_list):
        raise NotImplementedError(self._name)

    def rmsPoolCreate(self, config, rms_pool):
        raise NotImplementedError(self._name)

    def rmsPoolList(self, config):
        raise NotImplementedError(self._name)

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        raise NotImplementedError(self._name)

    def _autoDetectRMS(self, config):
        if config.get('rms', None) == self._name:
            return True

        return False

    def rmsProcessChecksums(self, config, rms_pool, package_list):
        ret = []

        for package in package_list:
            package = package.split('@', 1)
            filename = package[0]

            if len(package) == 2:
                hash_str = package[1]
                hash_type, hash = hash_str.split(':', 1)

                if hash_type not in self.ALLOWED_HASH_TYPES:
                    self._errorExit(
                        'Unsupported hash type "{0}"'.format(hash_type))

                self._info('Verifying {2} hash of {0} in {1}'.format(
                    filename, rms_pool, hash_type))
                rms_hash = self.rmsGetHash(
                    config, rms_pool, filename, hash_type)

                if rms_hash != hash:
                    self._errorExit(
                        'RMS hash mismatch "{0}" != "{1}"'.format(rms_hash, hash))

            ret.append(filename)

        return ret

    def rmsCalcHash(self, file_name, hash_type):
        hashlib = self._ext.hashlib

        hf = hashlib.new(hash_type)
        with open(file_name, 'rb') as f:
            for chunk in iter(lambda: f.read(65536), ''):
                if not chunk:
                    break
                hf.update(chunk)
        return "{0}:{1}".format(hash_type, hf.hexdigest())

    def rmsCalcHashes(self, file_name):
        hashlib = self._ext.hashlib

        hashes = {}

        for hash_type in self.ALLOWED_HASH_TYPES:
            hashes[hash_type] = hashlib.new(hash_type)

        with open(file_name, 'rb') as f:
            for chunk in iter(lambda: f.read(65536), ''):
                if not chunk:
                    break

                for hash_type in self.ALLOWED_HASH_TYPES:
                    hashes[hash_type].update(chunk)

        for hash_type in self.ALLOWED_HASH_TYPES:
            hashes[hash_type] = hashes[hash_type].hexdigest()

        return hashes



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn
from .curltoolmixin import CurlToolMixIn


class rustupTool(BashToolMixIn, CurlToolMixIn, RunEnvTool):
    """rustup is an installer for the systems programming language Rust.

Home: https://www.rustup.rs/
"""
    __slots__ = ()

    INSTALLER_DEFAULT = 'https://sh.rustup.rs'

    def getDeps(self):
        return (
            BashToolMixIn.getDeps(self) +
            CurlToolMixIn.getDeps(self))

    def _installTool(self, env):
        if self._detect.isAlpineLinux():
            self._warn('Unfortunately, rustup does not support musl libc yet')

        installer = self._callCurl(env, [env['rustupInstaller']])

        self._callBash(
            env,
            bash_args=['--', '-y', '--no-modify-path'],
            input=installer)

    def _updateTool(self, env):
        self._executil.callExternal([
            env['rustupBin'], 'self', 'update'
        ])

    def uninstallTool(self, env):
        ospath = self._ospath

        for v in ['rustupDir', 'cargoDir']:
            dir = env[v]
            self._pathutil.rmTree(dir)

        self._have_tool = False

    def envNames(self):
        return ['rustupBin', 'rustupDir', 'rustupInstaller']

    def initEnv(self, env):
        ospath = self._ospath
        environ = self._environ
        dir = ospath.join(environ['HOME'], '.multirust')
        dir = environ.setdefault('RUSTUP_HOME', dir)
        cargo_dir = ospath.join(environ['HOME'], '.cargo')
        cargo_dir = environ.setdefault('CARGO_HOME', cargo_dir)

        dir = env.setdefault('rustupDir', dir)
        cargo_dir = env.setdefault('cargoDir', cargo_dir)

        environ['RUSTUP_HOME'] = dir
        environ['CARGO_HOME'] = cargo_dir

        env.setdefault('rustupInstaller', self.INSTALLER_DEFAULT)

        bin_dir = ospath.join(cargo_dir, 'bin')
        self._pathutil.addBinPath(bin_dir, True)

        if ospath.exists(ospath.join(bin_dir, 'rustup')):
            super(rustupTool, self).initEnv(env)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class yarnTool(NpmToolMixIn, BuildTool):
    """YARN: fast, reliable, and secure dependency management.

Home: https://yarnpkg.com

Note: auto-detected only if yarn.lock is present
"""
    __slots__ = ()

    def getOrder(self):
        # required to run before other npm-based tools
        return -10

    def autoDetectFiles(self):
        return 'yarn.lock'

    def onPrepare(self, config):
        node_env = self._environ['NODE_ENV']

        try:
            self._environ['NODE_ENV'] = 'development'
            yarnBin = config['env']['yarnBin']
            self._executil.callExternal(
                [yarnBin, 'install', '--production=false'])
        finally:
            self._environ['NODE_ENV'] = node_env

    def onPackage(self, config):
        yarnBin = config['env']['yarnBin']
        cmd = [yarnBin, 'install', '--production']
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..vcstool import VcsTool
from ..rmstool import RmsTool


class svnTool(VcsTool, RmsTool):
    """Apache Subversion: Enterprise-class centralized version control for the masses.

Home: https://subversion.apache.org/
"""
    __slots__ = ()

    def _installTool(self, env):
        if self._detect.isSLES():
            self._install.zypperRepo(
                None,
                'http://download.opensuse.org/repositories/devel:tools:scm:svn'
                '/SLE_12_SP2/devel:tools:scm:svn.repo')

        self._install.debrpm(['subversion'])
        self._install.emerge(['dev-vcs/subversion'])
        self._install.pacman(['subversion'])
        self._install.apk(['subversion'])
        self._install.brew('subversion')

    def autoDetectFiles(self):
        return '.svn'

    def autoDetect(self, config):
        return (VcsTool.autoDetect(self, config) or
                RmsTool.autoDetect(self, config))

    def getDeps(self):
        return ['ssh']

    def vcsGetRepo(self, config, wc_dir=None):
        wc_dir = wc_dir or self._os.getcwd()

        svn_info = self._executil.callExternal([
            config['env']['svnBin'], 'info', '--xml', wc_dir
        ], verbose=False)

        svn_info = self._ext.minidom.parseString(svn_info)
        svn_info = svn_info.getElementsByTagName('url')
        url = svn_info[0].firstChild.nodeValue

        url = self._ext.re.sub('/(trunk|branches|tags).*$', '', url)
        return url

    def _detectSVNPath(self, config, vcs_ref):
        env = config['env']
        vcsRepo = config['vcsRepo']
        branch_path = '{0}/branches/{1}'.format(vcsRepo, vcs_ref)
        tag_path = '{0}/tags/{1}'.format(vcsRepo, vcs_ref)

        if vcs_ref == 'trunk':
            svn_repo_path = '{0}/trunk'.format(vcsRepo)
        elif self._callVCSSVN(config, ['info', branch_path], suppress_fail=True):
            svn_repo_path = branch_path
        elif self._callVCSSVN(config, ['info', tag_path], suppress_fail=True):
            svn_repo_path = tag_path
        else:
            self._errorExit("VCS ref was not found: " + vcs_ref)

        return svn_repo_path

    def vcsCheckout(self, config, vcs_ref):
        env = config['env']
        wc_dir = self._os.getcwd()
        vcs_ref = vcs_ref or 'trunk'

        svn_repo_path = self._detectSVNPath(config, vcs_ref)

        if self._ospath.isdir('.svn'):
            self._callVCSSVN(config, ['switch', svn_repo_path])
        else:
            self._callVCSSVN(config, [
                'checkout',
                svn_repo_path,
                wc_dir]
            )

    def vcsCommit(self, config, message, files):
        if files:
            commit_files = files
        else:
            commit_files = []
            files = ['--depth=infinity', '.']

        self._callVCSSVN(config, ['add', '--force', ] + files)
        self._callVCSSVN(config, ['commit', '-m', message] + commit_files)

    def vcsTag(self, config, tag, message):
        env = config['env']

        svn_info = self._callVCSSVN(config, ['info', '--xml'], verbose=False)
        svn_info = self._ext.minidom.parseString(svn_info)
        svn_info = svn_info.getElementsByTagName('url')
        svn_url = svn_info[0].firstChild.nodeValue

        self._callVCSSVN(config, [
            'copy',
            '-m', message,
            '--parents',
            svn_url,
            '%s/tags/%s' % (config['vcsRepo'], tag)
        ])

    def vcsPush(self, config, refs):
        pass

    def vcsGetRevision(self, config):
        svn_info = self._callVCSSVN(config, ['info',  '--xml'], verbose=False)

        svn_info = self._ext.minidom.parseString(svn_info)
        svn_info = svn_info.getElementsByTagName('commit')

        if len(svn_info):
            return svn_info[0].getAttribute('revision')

        return 'local'

    def vcsGetRefRevision(self, config, vcs_cache_dir, branch):
        svn_repo_path = self._detectSVNPath(config, branch)

        res = self._callVCSSVN(
            config, ['info', svn_repo_path, '--xml'], verbose=False)

        res = self._ext.minidom.parseString(res)
        return res.getElementsByTagName('commit')[0].getAttribute('revision')

    def vcsListTags(self, config, vcs_cache_dir, tag_hint):
        return self._svnListCommon(config, 'vcsRepo', vcs_cache_dir, 'tags')

    def vcsListBranches(self, config, vcs_cache_dir, branch_hint):
        return self._svnListCommon(config, 'vcsRepo', vcs_cache_dir, 'branches')

    def _svnListCommon(self, config, repo_key, vcs_cache_dir, sub_path):
        repo = config[repo_key]
        svn_repo_path = '{0}/{1}'.format(repo, sub_path)

        if svn_repo_path[-1] != '/':
            svn_repo_path += '/'

        res = self._callSVN(
            config, repo_key, ['ls', svn_repo_path],
            verbose=False
        )

        res = res.strip().split("\n")

        res = [v and v.replace('/', '') or '' for v in res]
        res = list(filter(None, res))
        return res

    def vcsExport(self, config, vcs_cache_dir, vcs_ref, dst_path):
        ospath = self._ospath
        svn_repo_path = self._detectSVNPath(config, vcs_ref)

        self._pathutil.rmTree(dst_path)

        if vcs_cache_dir is None:
            self._callVCSSVN(config, ['export', svn_repo_path, dst_path])
            return

        if ospath.isdir(vcs_cache_dir):
            cnd = 'switch'
        else:
            cnd = 'checkout'

        self._callVCSSVN(config, [cnd, svn_repo_path, vcs_cache_dir])
        self._callVCSSVN(config, ['export', vcs_cache_dir, dst_path])

    def vcsBranch(self, config, vcs_ref):
        vcsRepo = config['vcsRepo']
        svn_dst_path = '{0}/branches/{1}'.format(vcsRepo, vcs_ref)

        svn_info = self._callVCSSVN(config, ['info', '--xml'], verbose=False)
        svn_info = self._ext.minidom.parseString(svn_info)
        svn_info = svn_info.getElementsByTagName('url')
        svn_src_path = svn_info[0].firstChild.nodeValue

        if self._callVCSSVN(config, ['info', svn_dst_path], suppress_fail=True):
            self._errorExit('VCS branch {0} already exists!'.format(vcs_ref))

        self._callVCSSVN(config, [
            'copy', '--parents',
            '-m', 'CID branch ' + vcs_ref,
            svn_src_path, svn_dst_path,
        ])
        self.vcsCheckout(config, vcs_ref)

    def vcsMerge(self, config, vcs_ref, cleanup):
        svn_repo_path = self._detectSVNPath(config, vcs_ref)

        self._callVCSSVN(config, ['update'])
        self._callVCSSVN(
            config, ['merge', '--accept', 'postpone', svn_repo_path])

        try:
            self._callVCSSVN(config, ['commit', '-m', 'CID merged ' + vcs_ref])
        except self._ext.subprocess.CalledProcessError:
            if cleanup:
                self.vcsRevert(config)
                self._errorExit('Merge failed, aborted.')
            self._errorExit('Merge failed, left as-is.')

    def vcsDelete(self, config, vcs_cache_dir, vcs_ref):
        vcsRepo = config['vcsRepo']
        svn_repo_path = '{0}/branches/{1}'.format(vcsRepo, vcs_ref)
        self._callVCSSVN(
            config, ['remove', '-m', 'CID delete ' + vcs_ref, svn_repo_path])

    def vcsRevert(self, config):
        self._callVCSSVN(config, ['revert', '-R', '.'])

    def vcsIsMerged(self, config, vcs_ref):
        vcsRepo = config['vcsRepo']
        svn_repo_path = '{0}/branches/{1}'.format(vcsRepo, vcs_ref)

        res = self._callVCSSVN(config, [
            'mergeinfo', '--show-revs',
            'eligible',
            svn_repo_path
        ], verbose=False).strip()

        return res == ''

    def vcsClean(self, config):
        svn_info = self._executil.callExternal([
            config['env']['svnBin'], 'status', '--no-ignore', '--xml'
        ], verbose=False)

        svn_info = self._ext.minidom.parseString(svn_info)
        entries = svn_info.getElementsByTagName('entry')
        pathutil = self._pathutil

        for entry in entries:
            p = entry.getAttribute('path')
            wcstatus = entry.getElementsByTagName('wc-status')

            if wcstatus[0].getAttribute('item') in ('unversioned', 'ignored'):
                pathutil.rmTree(p)

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath
        rms_repo = config['rmsRepo']

        for package in package_list:
            package_basename = ospath.basename(package)

            dst = '{0}/{1}/{2}'.format(rms_repo, rms_pool, package_basename)

            self._callRMSSVN(config, [
                'import',
                '-m', 'FutoIn CID upload',
                package, dst,
            ])

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        ospath = self._ospath
        rms_repo = config['rmsRepo']

        args = []

        if '/' in dst_pool:
            args += ['--parents']

        for package in package_list:
            package_basename = ospath.basename(package)

            src = '{0}/{1}/{2}'.format(rms_repo, src_pool, package_basename)
            dst = '{0}/{1}/{2}'.format(rms_repo, dst_pool, package_basename)

            self._callRMSSVN(config, [
                'copy',
                '-m', 'FutoIn CID promotion',
                src, dst,
            ] + args)

    def rmsGetList(self, config, rms_pool, package_hint):
        return self._svnListCommon(config, 'rmsRepo', None, rms_pool)

    def rmsRetrieve(self, config, rms_pool, package_list):
        self._rmsRetrieve(config, rms_pool, package_list)

    def _rmsRetrieve(self, config, rms_pool, package_list, dst_dir=None):
        ospath = self._ospath
        rms_repo = config['rmsRepo']

        for package in package_list:
            package_basename = ospath.basename(package)

            src = '{0}/{1}/{2}'.format(rms_repo, rms_pool, package_basename)

            if dst_dir:
                dst = ospath.join(dst_dir, package_basename)
            else:
                dst = package_basename

            self._callRMSSVN(config, [
                'export',
                src, package_basename,
            ])

    def rmsPoolCreate(self, config, rms_pool):
        rms_repo = config['rmsRepo']

        dst = '{0}/{1}'.format(rms_repo, rms_pool)

        try:
            self._callRMSSVN(config, [
                'mkdir', '--parents',
                '-m', 'FutoIn CID pool creation',
                dst,
            ])
        except self._ext.subprocess.CalledProcessError:
            # make sure it exists
            self._callRMSSVN(config, ['info', dst])

    def rmsPoolList(self, config):
        return self._svnListCommon(config, 'rmsRepo', None, '')

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        rms_repo = config['rmsRepo']

        src = '{0}/{1}/{2}'.format(rms_repo, rms_pool, package)

        hf = self._ext.hashlib.new(hash_type)

        self._callRMSSVN(config, ['cat', src],
                         output_handler=lambda chunk: hf.update(chunk))

        return hf.hexdigest()

    def _callVCSSVN(self, config, args, **kwargs):
        return self._callSVN(config, 'vcsRepo', args, **kwargs)

    def _callRMSSVN(self, config, args, **kwargs):
        return self._callSVN(config, 'rmsRepo', args, **kwargs)

    def _callSVN(self, config, repo_key, args, **kwargs):
        env = config['env']
        svnBin = env['svnBin']
        rms_repo = config[repo_key]
        opts = []

        if rms_repo.startswith('svn+ssh://'):
            sshm = self._ext.re.match('svn\\+ssh://[^/]+(:([0-9]+))', rms_repo)
            ssh_port = sshm.group(2) or '22'
            ssh_opts = '-o BatchMode=yes -o StrictHostKeyChecking={0}'.format(
                env['sshStrictHostKeyChecking'])
            opts += [
                '--config-option',
                'config:tunnels:ssh=$SVN_SSH ssh -q -p {0} {1}'.format(
                    ssh_port, ssh_opts)
            ]

            # Dirty workaround
            if sshm.group(1):
                for i in range(len(args)):
                    a = args[i]

                    if a.startswith(sshm.group(0)):
                        args[i] = a.replace(sshm.group(1), '', 1)

        return self._executil.callExternal([svnBin] + args + opts, **kwargs)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from ..testtool import TestTool
from .sdkmantoolmixin import SdkmanToolMixIn


class mavenTool(SdkmanToolMixIn, BuildTool, TestTool):
    """Apache Maven is a software project management and comprehension tool.

Home: https://maven.apache.org/

Expects clean, compile, package and test targets.

Build targets:
    prepare -> clean
    build -> compile
    package -> package
    check -> test
Override targets with .config.toolTune.

Requires Java >= 7.
"""
    __slots__ = ()

    def _minJava(self):
        return '7'

    def autoDetectFiles(self):
        return 'pom.xml'

    def _binName(self):
        return 'mvn'

    def _callMaven(self, config, target):
        cmd = [config['env']['mavenBin'], target]
        self._executil.callMeaningful(cmd)

    def onPrepare(self, config):
        target = self._getTune(config, 'prepare', 'clean')
        self._callMaven(config, target)

    def onBuild(self, config):
        target = self._getTune(config, 'build', 'compile')
        self._callMaven(config, target)

    def onPackage(self, config):
        target = self._getTune(config, 'package', 'package')
        self._callMaven(config, target)
        self._pathutil.addPackageFiles(config, 'target/*.jar')

    def onCheck(self, config):
        target = self._getTune(config, 'check', 'test')
        self._callMaven(config, target)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class gpgTool(RunEnvTool):
    """The GNU Privacy Guard.

Home: https://www.gnupg.org/

gpgKeyServer is hkp://keyserver.ubuntu.com:80 by default
"""
    __slots__ = ()

    def envNames(self):
        return ['gpgBin', 'gpgKeyServer']

    def _installTool(self, env):
        self._install.debrpm('gnupg')
        self._install.debrpm('gnupg2')
        self._install.debrpm('dirmngr')

        self._install.emerge(['app-crypt/gnupg'])
        self._install.pacman(['gnupg'])
        self._install.apk(['gnupg'])
        self._install.brew('gnupg')

    def initEnv(self, env):
        super(gpgTool, self).initEnv(env)
        env.setdefault('gpgKeyServer', 'hkp://keyserver.ubuntu.com:80')

        if self._have_tool and self._detect.isDeb():
            self._have_tool = self._pathutil.which('dirmngr')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from ..testtool import TestTool
from ..rmstool import RmsTool


class cargoTool(BuildTool, TestTool, RmsTool):
    """Cargo, Rust;s Package Manager.

Home: http://doc.crates.io/

Build targets:
    prepare -> clean
    build -> build
    check -> test
Override targets with .config.toolTune.

"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'Cargo.toml'

    def getDeps(self):
        return ['rust']

    def _installTool(self, env):
        if self._detect.isAlpineLinux():
            self._install.apk('cargo')

    def uninstallTool(self, env):
        pass

    def onPrepare(self, config):
        self._executil.callExternal([
            config['env']['cargoBin'], 'clean',
        ])

    def onBuild(self, config):
        args = []

        if not config.get('debugBuild', False):
            args.append('--release')

        self._executil.callExternal(
            [config['env']['cargoBin'], 'build'] + args)

    def onPackage(self, config):
        cmd = [config['env']['cargoBin'], 'package', '--allow-dirty']
        self._executil.callMeaningful(cmd)
        self._pathutil.addPackageFiles(config, 'target/package/*.crate')

    def onCheck(self, config):
        cmd = [config['env']['cargoBin'], 'test']
        self._executil.callMeaningful(cmd)

    def onRunDev(self, config):
        cmd = [config['env']['cargoBin'], 'run']
        self._executil.callMeaningful(cmd)

    def rmsUpload(self, config, rms_pool, package_list):
        cmd = [config['env']['cargoBin'], 'publish']
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn
from .curltoolmixin import CurlToolMixIn


class sdkmanTool(BashToolMixIn, CurlToolMixIn, RunEnvTool):
    """SDK Man for Java.

Home: http://sdkman.io/
"""
    __slots__ = ()

    def getDeps(self):
        return (
            ['unzip', 'zip'] +
            BashToolMixIn.getDeps(self) +
            CurlToolMixIn.getDeps(self))

    def envNames(self):
        return ['sdkmanDir', 'sdkmanGet']

    def _installTool(self, env):
        dir = env['sdkmanDir']
        get = env.get('sdkmanGet', 'https://get.sdkman.io')

        installer = self._callCurl(env, [get])

        environ = self._environ
        environ['SDKMAN_DIR'] = dir
        self._callBash(env, input=installer)
        del environ['SDKMAN_DIR']

    def _updateTool(self, env):
        self._callBash(env,
                       'source {0} && sdk selfupgrade'.format(
                           env['sdkmanInit'])
                       )

    def uninstallTool(self, env):
        dir = env['sdkmanDir']
        self._pathutil.rmTree(dir)
        self._have_tool = False

    def initEnv(self, env):
        ospath = self._ospath
        dir = ospath.join(self._environ['HOME'], '.sdkman')
        dir = env.setdefault('sdkmanDir', dir)
        env_init = ospath.join(dir, 'bin', 'sdkman-init.sh')
        env['sdkmanInit'] = env_init
        self._have_tool = ospath.exists(env_init)

    def onExec(self, env, args, replace=True):
        cmd = '. {0} && sdk {1}'.format(
            env['sdkmanInit'], self._ext.subprocess.list2cmdline(args))
        self._callBashInteractive(env, cmd, replace=replace)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class JavaToolMixIn(object):
    __slots__ = ()

    _LATEST_JAVA = '8'

    def _minJava(self):
        return None

    def envDeps(self, env):
        min_java = self._minJava()

        if not min_java:
            return

        for ek in ['javaVer', 'jdkVer']:
            if env.get(ek, None):
                if int(env[ek]) < int(min_java):
                    self._errorExit('{0} requires minimal {1}={2}'
                                    .format(self._name, ek, min_java))
            else:
                env[ek] = str(self._LATEST_JAVA)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext
from . import simple_memo as _simple_memo
from . import complex_memo as _complex_memo
from . import log as _log

_SYSTEM_VER = 'system'


@_simple_memo
def binaryVersions():
    detect = _ext.detect

    if detect.isDebian() or detect.isUbuntu():
        return ['5.6', '7.0', '7.1', '7.2']

    if detect.isSCLSupported():
        return ['5.6', '7.0', '7.1', '7.2']

    if detect.isArchLinux():
        # return ['5.6', '7.0', '7.1']
        return None

    if detect.isAlpineLinux():
        alpine_ver = detect.alpineLinuxVersion()

        if alpine_ver >= ['3', '8']:
            return ['5.6', '7.2']

        if alpine_ver >= ['3', '6']:
            return ['5.6', '7.1']

        if alpine_ver >= ['3', '5']:
            return ['5.6', '7.0']

        return ['5.6']

    if detect.isMacOS():
        return ['5.6', '7.0', '7.1', '7.2']

    return None


@_complex_memo
def basePackage(ver):
    detect = _ext.detect
    ver_nodot = ver.replace('.', '')

    if detect.isDebian() or detect.isUbuntu():
        if ver == _SYSTEM_VER:
            if detect.osCodeName() in ['jessie', 'trusty']:
                return 'php5'

            return 'php'

        return 'php{0}'.format(ver)

    if detect.isSCLSupported():
        if isIUSVer(ver):
            return 'php{0}u'.format(ver_nodot)
        else:
            return 'rh-php{0}'.format(ver_nodot)

    if detect.isArchLinux():
        if isArchLatest(ver):
            return 'php'

        return 'php{0}'.format(ver_nodot)

    if detect.isAlpineLinux():
        if isAlpineSplit():
            return 'php{0}'.format(ver.split('.')[0])

        return 'php'

    if detect.isMacOS():
        return 'homebrew/php/php{0}'.format(ver_nodot)

    if detect.isFedora():
        return 'php'

    if detect.isOpenSUSE():
        return 'php7'

    if detect.isSLES():
        return 'php7'

    return 'php'


def createBinDir(env, src_bin):
    ospath = _ext.ospath
    os = _ext.os
    pathutil = _ext.pathutil

    php_dir = ospath.join(os.environ['HOME'], '.php', env['phpVer'])
    php_dir = env.setdefault('phpDir', php_dir)

    php_bin_dir = ospath.join(php_dir, 'bin')
    php_bin = ospath.join(php_bin_dir, 'php')

    if not ospath.exists(php_bin_dir):
        os.makedirs(php_bin_dir)

    _ext.pathutil.addBinPath(php_bin_dir, True)

    if ospath.exists(src_bin):
        if ospath.islink(php_bin) and os.readlink(php_bin) == src_bin:
            pass
        else:
            pathutil.rmTree(php_bin)
            os.symlink(src_bin, php_bin)

        env['phpBin'] = php_bin
        env['phpFakeBinDir'] = php_bin_dir
        return True

    elif 'phpBin' in env:
        del env['phpBin']

    return False


@_simple_memo
def isAlpineSplit():
    return _ext.detect.alpineLinuxVersion() >= ['3', '5']


def isArchLatest(ver):
    # return ver == '7.1'
    return True


def isIUSVer(ver):
    return ver == '7.1'


def installedExtensions(env):
    known = {k: None for k in knownExtensions()}
    #---
    res = _ext.executil.callExternal(
        [env['phpBin'], '-m'],
        verbose=False)
    res = res.strip().split('\n')

    for r in res:
        r = r.lower().strip()

        if r == 'Zend OPcache':
            r = 'opcache'

        if r == 'mysqlnd':
            r = 'mysql'

        if r in known:
            known[r] = True

    return known


def extPackages(env, known=None):  # NOTE: do not cache
    ver = env['phpVer']
    base_pkg = basePackage(ver)
    pkg_prefix = '{0}-'.format(base_pkg)
    pkg_prefix2 = None
    detect = _ext.detect
    install = _ext.install

    known = known or installedExtensions(env)
    update = {}
    pkg2key = {}
    found = []

    #---
    if detect.isDebian() or detect.isUbuntu():
        pkg2key = {
            'mysqlnd': 'mysql',
        }
        pkg_prefix2 = 'php-'

    elif detect.isSCLSupported():
        if isIUSVer(ver):
            pkg_prefix2 = '{0}-pecl-'.format(base_pkg)
        else:
            pkg_prefix = '{0}-php-'.format(base_pkg)
            pkg_prefix2 = '{0}-php-pecl-'.format(base_pkg)

        pkg2key = {
            'mysqlnd': 'mysql',
            'process': 'pcntl',
        }

    elif detect.isArchLinux() or detect.isGentoo():
        pkg_prefix = None

    elif detect.isMacOS():
        pkg2key = {
            'pdo-pgsql': 'pgsql',
        }

    elif detect.isFedora():
        pkg_prefix2 = pkg_prefix + 'pecl-'
        pkg2key = {
            'mysqlnd': 'mysql',
            'process': 'pcntl',
        }

    #---
    if pkg_prefix:
        found += install.search(pkg_prefix)

    if pkg_prefix2:
        found += install.search(pkg_prefix2)

    #---
    if found:
        for p in found:
            # NOTE: in that order!
            if pkg_prefix2:
                k = p.replace(pkg_prefix2, '')

            k = k.replace(pkg_prefix, '')

            if not k:
                continue

            k = pkg2key.get(k, k)

            if k in known:
                update[k] = p

    #---
    for (k, v) in update.items():
        if k in known:
            if not known[k]:
                known[k] = v
        else:
            _log.errorExit('Unknown generic PHP ext "{0}"'.format(k))

    return known


def knownExtensions():
    return [
        'amqp',
        'apcu',
        'ast',
        'bcmath',
        'bz2',
        'blitz',
        'calendar',
        'ctype',
        'curl',
        'couchbase',
        'date',
        'ds',
        'ev',
        'enchant',
        'event',
        'exif',
        'fileinfo',
        'filter',
        'ftp',
        'gettext',
        'gd',
        'gearman',
        'geoip',
        'gmagick',
        'gmp',
        'gnupg',
        'grpc',
        'hash',
        'http',
        'iconv',
        'imagick',
        'imap',
        'intl',
        'json',
        'ldap',
        'kafka',
        'mailparse',
        'mbstring',
        'mysql',
        'mcrypt',
        'memcache',
        'memcached',
        'mongodb',
        'msgpack',
        'odbc',
        'opcache',
        'openssl',
        'pcntl',
        'pcre',
        'pdo',
        'pdo_mysql',
        'pdo_pgsql',
        'pdo_sqlite',
        'phar',
        'pgsql',
        'posix',
        'pspell',
        'readline',
        'recode',
        'redis',
        'session',
        'shmop',
        'snmp',
        'simplexml',
        'soap',
        'sockets',
        'spl',
        'sqlite3',
        'ssh2',
        'sysvmsg',
        'sysvsem',
        'sysvshm',
        'tokenizer',
        'uv',
        'xdebug',
        'xml',
        'xmlreader',
        'xmlrpc',
        'xmlwriter',
        'xsl',
        'zlib',
        'zip',
        'zmq',
    ]


def installExtensions(env, exts, permissive=True):
    exts = _ext.configutil.listify(exts)

    #---
    known = installedExtensions(env)

    for ext in exts:
        if ext in known:
            if known[ext] != True:
                break
        else:
            _log.errorExit('Unknown PHP extension "{0}"\nKnown: \n* {1}'.format(
                ext, '\n* '.join(knownExtensions())))
    else:
        return

    #---
    mapping = extPackages(env, known)
    install = _ext.install

    to_install = []

    for ext in exts:
        if ext in mapping:
            pkg = mapping[ext]

            if type(pkg) == type(''):
                to_install.append(pkg)
            elif not pkg:
                msg = 'Not supported PHP extension: {0}'.format(ext)

                if permissive:
                    _log.warn(msg)
                else:
                    _log.errorExit(msg)
        else:
            _log.errorExit('Unknown PHP extension "{0}"\nKnown: \n* {1}'.format(
                ext, '\n* '.join(knownExtensions())))

    #---
    detect = _ext.detect

    if detect.isDisabledToolsSetup(env):
        _log.errorExit(
            'environment requires external management of tools')
        return

    if detect.isExternalToolsSetup(env):
        _ext.executil.externalSetup(env, [
            'tool', 'install', 'php'])
        return

    #---
    for pkg in to_install:
        install.deb(pkg)
        install.rpm(pkg)
        install.brew(pkg)
        install.apk(pkg)
        install.pacman(pkg)
        install.emerge(pkg)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..rmstool import RmsTool


class nexus3Tool(RmsTool):
    """Sonatype Nexus RMS.
Home: http://www.sonatype.org/nexus/

PLEASE DO NOT USE. INCOMPLETE.

Nexus 3 does not support REST API yet.
So, this tool is not usable yet.

Please set user and password in ~/.futoin.json env. Please
avoid setting these variables in process environment.
"""
    __slots__ = ()

    def envNames(self):
        return ['nexus3User', 'nexus3Password']

    def initEnv(self, env):
        self._have_tool = True

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath

        for package in package_list:
            package_basename = ospath.basename(package)

            with open(package, 'rb') as pf:
                self._callNexus(
                    config,
                    'PUT',
                    '/repository/{0}/{1}'.format(rms_pool, package_basename),
                    data=pf
                )

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        # TODO: use Groovy API to transfer on server
        self.rmsRetrieve(config, src_pool, package_list)
        self.rmsUpload(config, dst_pool, package_list)

    def rmsGetList(self, config, rms_pool, package_hint):
        """script_name = 'cid_list'.format(
            rms_pool.replace('/', '$'),
            package_hint.replace('?', '$$').replace('*', '$$$')
        )

        script = [
            'return groovy.json.JsonOutput.toJson(',
        ]

        self._execNexusScript(config, script_name, script)
        """
        raise NotImplementedError(self._name)

    def rmsRetrieve(self, config, rms_pool, package_list):
        for package in package_list:
            result = self._callNexus(
                config,
                'GET',
                '/repository/{0}/{1}'.format(rms_pool, package),
                stream=True
            )
            result.raise_for_status()

            with open(package, 'wb') as f:
                result.raw.decode_content = True
                self._ext.shutil.copyfileobj(result.raw, f)

    def rmsPoolCreate(self, config, rms_pool):
        pool = rms_pool.split('/', 1)[0]

        script_name = 'cid_pool_create'

        script = [
            'import groovy.json.JsonSlurper'
            'def pool = new JsonSlurper().parseText(args)'
            "def rawStore = blobStore.createFileBlobStore(pool.name, pool.path)",
            "repository.createRawHosted(pool.name, rawStore.name)",
        ]

        self._execNexusScript(config, script_name, script, json={
            'name': pool,
            'path': pool,
        })

    def rmsPoolList(self, config):
        raise NotImplementedError(self._name)

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        raise NotImplementedError(self._name)

    def _execNexusScript(self, config, script_name, script, **kwargs):
        result = self._callNexus(
            config,
            'PUT',
            '/service/siesta/rest/v1/script',
            json={
                'name': script_name,
                'type': 'groovy',
                'content': "\n".join(script),
            }
        )

        result.raise_for_status()

        result = self._callNexus(
            config,
            'POST',
            '/service/siesta/rest/v1/script/{0}/run'.format(script_name),
            **kwargs
        )

        result.raise_for_status()

        return result

    def _callNexus(self, config, method, path, **kwargs):
        env = config['env']

        rms_repo = config['rmsRepo']
        if rms_repo[-1] == '/':
            path = path[1:]

        url = rms_repo + path

        if 'nexus3User' in env and 'nexus3Password' in env:
            kwargs['auth'] = (env['nexus3User'], env['nexus3Password'])

        self._configutil.requestsOptions(env, kwargs)

        self._info('HTTP call {0} {1}'.format(method, url))
        return self._ext.requests.request(method, url, **kwargs)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .javatoolmixin import JavaToolMixIn


class jdkTool(RunEnvTool, JavaToolMixIn):
    """Java Development Kit.

Home: http://openjdk.java.net/

Due to issues with Oracle's licensing, cid
supports only automatic installation of OpenJDK.

jdkVer supports:
- only one digit like 6, 7, 8, 9 (Zulu JDK (http://zulu.org)
- use jdkDir for custom JDK
- jdkVer is equal to javaVer by default
"""
    __slots__ = ()

    def getDeps(self):
        return ['java']

    def envNames(self):
        return ['jdkDir', 'jdkVer']

    def getVersionParts(self):
        return 1

    def _installTool(self, env):
        if env.get('jdkDir', None):
            return

        ver = env['jdkVer']

        # Zulu is installed in javaTool
        # leaving it here for possible future use
        # self._install.deb(['openjdk-{0}-jdk'.format(ver)])
        # self._install.yum(['java-1.{0}.0-openjdk-devel'.format(ver)])
        # self._install.zypper(['java-1_{0}_0-openjdk-devel'.format(ver)])
        self._install.pacman(['jdk{0}-openjdk'.format(ver)])
        self._install.emerge(['=dev-java/oracle-jdk-bin-1.{0}*'.format(ver)])

        if self._detect.isAlpineLinux():
            self._install.apkCommunity()
            self._install.apk('openjdk{0}'.format(ver))
            return

    def uninstallTool(self, env):
        pass

    def initEnv(self, env):
        ospath = self._ospath
        environ = self._environ
        glob = self._ext.glob
        detect = self._detect

        if env.get('jdkDir', None):
            java_home = env['jdkDir']
            bin_dir = ospath.join(java_home, 'bin')

            environ['JAVA_HOME'] = java_home
            environ['JDK_HOME'] = java_home

            self._pathutil.addBinPath(bin_dir, True)

            super(jdkTool, self).initEnv(env, 'javac')
            return

        env.setdefault('jdkVer', env['javaVer'])
        ver = env['jdkVer']

        candidates = [
            # Zulu
            '/usr/lib/jvm/zulu-{0}*/bin/javac'.format(ver),
            # Debian / Ubuntu
            #"/usr/lib/jvm/java-{0}-openjdk*/bin/javac".format(ver),
            # RedHat
            #"/usr/lib/jvm/java-1.{0}.0/bin/javac".format(ver),
            # OpenSuse
            #"/usr/lib*/jvm/java-1.{0}.0/bin/javac".format(ver),
            # Default oracle
            #"/opt/jdk/jdk1.{0}*/bin/javac".format(ver),
        ]

        if detect.isGentoo() or detect.isArchLinux():
            candidates += [
                "/usr/lib/jvm/java-{0}-openjdk*/bin/javac".format(ver),
            ]
        elif detect.isAlpineLinux():
            candidates += [
                "/usr/lib/jvm/java-1.{0}-openjdk/bin/javac".format(ver),
            ]
        elif detect.isMacOS():
            candidates += [
                "/Library/Java/JavaVirtualMachines/zulu-{0}.jdk/Contents/Home/bin/javac".format(
                    ver)
            ]

        for c in candidates:
            if ospath.exists(c):
                bin_name = [c]
            else:
                bin_name = glob.glob(c)

            if bin_name:
                bin_name = bin_name[0]
                bin_dir = ospath.dirname(bin_name)
                java_home = ospath.dirname(bin_dir)

                env['jdkDir'] = java_home
                environ['JAVA_HOME'] = java_home
                environ['JDK_HOME'] = java_home

                self._pathutil.addBinPath(bin_dir, True)
                self._have_tool = True
                break



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool


class releaseTool(BuildTool):
    """FutoIn CID-specific release processing

Python:
- replaces "*__version__"' with "__version__ = '{version}'"
Changelog:
- replaces "=== \(next\) ===" with "=== {version} ({date}) ==="
- replaces "# \(next\)" with "# {version} ({date})"

Tune:
    .python = [] - list of Python files
    .changelog = 'CHANGELOG.txt' - list of ChangeLog files
"""
    __slots__ = ()

    def autoDetect(self, config):
        return True

    def updateProjectConfig(self, config, updates):
        re = self._ext.re

        def py_updater(content):
            if 'version' in updates:
                return re.sub(
                    r'^.*__version__.*$',
                    '__version__ = \'{0}\''.format(updates['version']),
                    content,
                    flags=re.MULTILINE
                )

        def cl_updater(content):
            if 'version' in updates:
                version = updates['version']
                date = self._ext.datetime.datetime.utcnow().isoformat().split('T')[
                    0]
                content = re.sub(
                    r'^=== \(next\) ===$',
                    '=== {0} ({1}) ==='.format(version, date),
                    content,
                    count=1,
                    flags=re.MULTILINE
                )
                content = re.sub(
                    r'# \(next\)',
                    '# {0} ({1})'.format(version, date),
                    content,
                    count=1,
                    flags=re.MULTILINE
                )
                return content

        res = []
        #---
        py_files = self._getTune(config, 'python', [])
        py_files = self._configutil.listify(py_files)

        for pyf in py_files:
            res += self._pathutil.updateTextFile(pyf, py_updater)

        #---
        cl_files = self._getTune(config, 'changelog', 'CHANGELOG.txt')
        cl_files = self._configutil.listify(cl_files)

        for clf in cl_files:
            res += self._pathutil.updateTextFile(clf, cl_updater)

        #---
        return res

    def initEnv(self, env):
        self._have_tool = True



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..rmstool import RmsTool


class nexusTool(RmsTool):
    """Sonatype Nexus RMS.
Home: http://www.sonatype.org/nexus/

Sonatype Nexus v2 (!). Please see nexus3 tools for v3 support.

Please set user and password in ~/.futoin.json env. Please
avoid setting these variables in process environment.
"""
    __slots__ = ()

    def envNames(self):
        return ['nexusUser', 'nexusPassword']

    def initEnv(self, env):
        self._have_tool = True

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath

        for package in package_list:
            package_basename = ospath.basename(package)
            path = '/content/repositories/{0}/{1}'.format(
                rms_pool, package_basename)

            if self._callNexus(config, 'HEAD', path).ok:
                self._errorExit(
                    'Package {0} already exists on RMS'.format(package_basename))

            with open(package, 'rb') as pf:
                res = self._callNexus(
                    config,
                    'PUT', path,
                    data=pf
                )
                res.raise_for_status()

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        # TODO: find out how to copy on server
        tmpdir = self._pathutil.tmpCacheDir(prefix='nexus')
        os = self._os

        try:
            cwd = os.getcwd()
            os.chdir(tmpdir)
            self.rmsRetrieve(config, src_pool, package_list)
            self.rmsUpload(config, dst_pool, package_list)
            os.chdir(cwd)
        finally:
            self._pathutil.rmTree(tmpdir)

    def rmsGetList(self, config, rms_pool, package_hint):
        rms_pool = rms_pool.split('/', 1)
        pool = rms_pool[0]
        path = rms_pool[1] if len(rms_pool) == 2 else ''

        res = self._callNexus(
            config,
            'GET', '/service/local/repositories/{0}/content/{1}'.format(
                pool, path),
            headers={'Accept': 'application/json'},
        )
        res.raise_for_status()
        res = res.json()

        res = [r['leaf'] and r['text'] or None for r in res['data']]
        res = list(filter(None, res))
        return res

    def rmsRetrieve(self, config, rms_pool, package_list):
        shutil = self._ext.shutil

        for package in package_list:
            result = self._callNexus(
                config,
                'GET',
                '/content/repositories/{0}/{1}'.format(rms_pool, package),
                stream=True
            )
            result.raise_for_status()

            with open(package, 'wb') as f:
                result.raw.decode_content = True
                shutil.copyfileobj(result.raw, f)

    def rmsPoolCreate(self, config, rms_pool):
        res = self._callNexus(
            config,
            'GET', '/service/local/repositories/{0}'.format(rms_pool),
        )

        if res.ok:
            return

        res = self._callNexus(
            config,
            'POST', '/service/local/repositories',
            json={
                'data': {
                    'id': rms_pool,
                    'name': rms_pool,
                    'provider': 'site',
                    'format': 'site',
                    'repoType': 'hosted',
                    'exposed': True,
                    'writePolicy': 'ALLOW_WRITE_ONCE',
                    'browseable': True,
                    "repoPolicy": "MIXED",
                    "providerRole": "org.sonatype.nexus.proxy.repository.WebSiteRepository",
                }
            }
        )
        res.raise_for_status()

    def rmsPoolList(self, config):
        res = self._callNexus(
            config,
            'GET', '/service/local/all_repositories',
            headers={'Accept': 'application/json'},
        )
        res.raise_for_status()
        res = res.json()

        # case with no repos
        if 'data' not in res:
            return []

        return [r['id'] for r in res['data']]

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        path = '{0}/{1}'.format(rms_pool, package)
        pool, path = path.split('/', 1)

        res = self._callNexus(
            config,
            'GET', '/service/local/repositories/{0}/content/{1}?describe=info'.format(
                pool, path),
            headers={'Accept': 'application/json'},
        )
        res.raise_for_status()
        res = res.json()

        return res['data'][hash_type + 'Hash']

    def _callNexus(self, config, method, path, **kwargs):
        env = config['env']

        rms_repo = config['rmsRepo']
        if rms_repo[-1] == '/':
            path = path[1:]

        url = rms_repo + path

        if 'nexusUser' in env and 'nexusPassword' in env:
            kwargs['auth'] = (env['nexusUser'], env['nexusPassword'])

        self._configutil.requestsOptions(env, kwargs)

        self._info('HTTP call {0} {1}'.format(method, url))
        return self._ext.requests.request(method, url, **kwargs)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool


class makeTool(BuildTool):
    """GNU Make.

Home: https://www.gnu.org/software/make/

Expects presence of "clean" target.
Build uses the default target.
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return [
            'GNUmakefile',
            'makefile',
            'Makefile',
        ]

    def _installTool(self, env):
        self._install.debrpm(['make'])
        self._install.emerge(['sys-devel/make'])
        self._install.pacman(['make'])
        self._install.apk(['make'])

    def onPrepare(self, config):
        cmd = [config['env']['makeBin'], 'clean']
        self._executil.callMeaningful(cmd)

    def onBuild(self, config):
        target = self._getTune(config, 'build')

        if target:
            args = [target]
        else:
            args = []

        cmd = [config['env']['makeBin']] + args
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .data import DataSlots


class TimeoutException(RuntimeError):
    __slots__ = ()

    @classmethod
    def alarmHandler(cls, *args, **kwargs):
        raise TimeoutException('Alarm')


class ServiceMixIn(DataSlots):
    __slots__ = ()

    __MASTER_PID_FILE = '.futoin.master.pid'
    __RESTART_DELAY = 10
    __RESTART_DELAY_THRESHOLD = __RESTART_DELAY * 2

    def _serviceAdapt(self):
        if not self._config['adaptDeploy']:
            return

        self._deployLock()
        try:
            self._deploy_setup()
        finally:
            self._deployUnlock()

    def _serviceList(self):
        self._deployLock()
        try:
            self._initConfig()
        finally:
            self._deployUnlock()

        return self._configServiceList(self._config)

    def _configServiceList(self, config):
        entry_points = config.get('entryPoints', {})
        auto_services = config.get('deploy', {}).get('autoServices', {})
        tool_tune = config.get('toolTune', {})
        res = []

        deepCopy = self._ext.copy.deepcopy
        deepMerge = self._configutil.deepMerge

        for (ep, ei) in entry_points.items():
            if ep not in auto_services:
                self._errorExit(
                    'Unknown service entry point "{0}"'.format(ep))

            i = 0

            for svctune in auto_services[ep]:
                r = deepCopy(ei)
                r['name'] = ep
                r['instanceId'] = i
                r_tune = deepCopy(tool_tune.get(r['tool'], {}))
                deepMerge(r_tune, r.get('tune', {}))
                deepMerge(r_tune, svctune)

                if r_tune.get('socketAddress', None) == '0.0.0.0':
                    r_tune['socketAddress'] = '127.0.0.1'

                r['tune'] = r_tune

                res.append(r)
                i += 1

        return res

    def _serviceListPrint(self):
        for svc in self._serviceList():
            svc_tune = svc['tune']

            #---
            internal = svc_tune.get('internal', False)

            if not internal:
                internal = not bool(svc_tune.get('socketType', False))

            if internal:
                print("{0}\t{1}".format(svc['name'], svc['instanceId']))
                continue

            #---
            socket_type = svc_tune['socketType']

            if socket_type == 'unix':
                socket_addr = svc_tune['socketPath']
            else:
                socket_addr = '{0}:{1}'.format(
                    svc_tune['socketAddress'],
                    svc_tune['socketPort']
                )

            print("\t".join([svc['name'], str(svc['instanceId']),
                             socket_type, socket_addr]))

    def _serviceCommon(self, entry_point, instance_id):
        config = self._config
        entry_points = config.get('entryPoints', {})
        auto_services = config.get('deploy', {}).get('autoServices', {})

        try:
            instance_id = int(instance_id)
        except ValueError:
            self._errorExit('Invalid instance ID "{0}"'.format(instance_id))

        try:
            ep = entry_points[entry_point]
        except KeyError:
            self._errorExit('Unknown entry point "{0}"'.format(entry_point))

        try:
            dep = auto_services[entry_point][instance_id]
        except KeyError:
            self._errorExit(
                'Unknown service entry point "{0}"'.format(entry_point))
        except IndexError:
            self._errorExit('Unknown service entry point "{0}" instance "{1}"'.format(
                entry_point, instance_id))

        res = self._ext.copy.deepcopy(ep)
        res.setdefault('tune', {}).update(dep)
        return res

    def _serviceStop(self, svc, toolImpl, pid):
        signal = self._ext.signal
        os = self._os
        from ..runtimetool import RuntimeTool

        signal.signal(signal.SIGALRM, TimeoutException.alarmHandler)

        tune = svc['tune']

        try:
            toolImpl.onStop(self._config, pid, tune)
        except Exception as e:
            self._warn(str(e))
            return

        try:
            timeout = tune.get(
                'exitTimeoutMS', RuntimeTool.DEFAULT_EXIT_TIMEOUT)
            timeout /= 1000.0
            signal.setitimer(signal.ITIMER_REAL, timeout)

            try:
                os.waitpid(pid, 0)
            except TimeoutException:
                signal.setitimer(signal.ITIMER_REAL, 0)
                os.kill(pid, signal.SIGKILL)
                os.waitpid(pid, 0)
        except OSError:
            pass
        finally:
            signal.setitimer(signal.ITIMER_REAL, 0)

    def _serviceMasterPID(self):
        self._requireDeployLock()

        if not self._ospath.exists(self.__MASTER_PID_FILE):
            return None

        try:
            pid = int(self._pathutil.readTextFile(self.__MASTER_PID_FILE))
        except ValueError:
            return None

        try:
            self._masterLock()
            self._masterUnlock()
            return None
        except:
            pass

        return pid

    def _serviceMaster(self):
        signal = self._ext.signal
        os = self._os
        sys = self._sys
        time = self._ext.time
        from ..runtimetool import RuntimeTool

        svc_list = []
        pid_to_svc = {}

        self._running = True
        self._reload_services = True
        self._interruptable = False

        def serviceExitSignal(*args, **kwargs):
            self._running = False
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            signal.signal(signal.SIGUSR1, signal.SIG_IGN)
            signal.signal(signal.SIGUSR2, signal.SIG_IGN)

            if self._interruptable:
                raise KeyboardInterrupt('Exit received')

        def serviceReloadSignal(*args, **kwargs):
            self._reload_services = True

            if self._interruptable:
                raise KeyboardInterrupt('Reload received')

        signal.signal(signal.SIGTERM, serviceExitSignal)
        signal.signal(signal.SIGINT, serviceExitSignal)
        signal.signal(signal.SIGHUP, serviceExitSignal)
        signal.signal(signal.SIGUSR1, serviceReloadSignal)
        signal.signal(signal.SIGUSR2, serviceReloadSignal)
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)  # important for macOS

        # Still, it's not safe to assume processes continue
        # to run in set process group.
        # DO NOT use os.killpg()
        #os.setpgid(0, 0)

        # Main loop
        #---
        self._deployLock()
        try:
            current_master = self._serviceMasterPID()

            if current_master:
                self._errorExit(
                    'Master process is already running with PID "{0}"'.format(current_master))

            self._masterLock()
            self._pathutil.writeTextFile(self.__MASTER_PID_FILE,
                                         '{0}'.format(os.getpid()))
        finally:
            self._deployUnlock()

        self._info('Starting master process')
        self._dumpResourceStats()

        while self._running:
            # Reload services
            #---
            if self._reload_services:
                self._info('Reloading services')

                # Mark shutdown by default
                for svc in svc_list:
                    svc['_remove'] = True

                # Prepare process list
                for newsvc in self._serviceList():
                    for svc in svc_list:
                        for (sk, sv) in newsvc.items():
                            if svc.get(sk, None) != sv:
                                break
                        else:
                            break
                    else:
                        svc = newsvc
                        svc_list.append(newsvc)
                        svc['_pid'] = None
                        svc['_lastExit1'] = self.__RESTART_DELAY_THRESHOLD + 1
                        svc['_lastExit2'] = 0

                        tool = svc['tool']
                        t = self._getTool(tool)

                        if not t.isInstalled(self._config['env']):
                            self._errorExit(
                                'Tool "{0}" for "{1}" is not installed'
                                .format(tool, newsvc['name']))
                        elif isinstance(t, RuntimeTool):
                            newsvc['toolImpl'] = t
                        else:
                            self._errorExit(
                                'Tool "{0}" for "{1}" is not of RuntimeTool type'
                                .format(tool, newsvc['name']))

                        self._info('Added "{0}:{1}"'.format(
                            svc['name'], svc['instanceId']))

                    svc['_remove'] = False

                if not len(svc_list):
                    break

                # Kill removed or changed services
                for svc in svc_list:
                    if svc['_remove']:
                        pid = svc['_pid']

                        if pid:
                            self._serviceStop(svc, svc['toolImpl'], pid)
                            pid_to_svc[pid]['_pid'] = None
                            del pid_to_svc[pid]

                        self._info('Removed "{0}:{1}" pid "{2}"'.format(
                            svc['name'], svc['instanceId'], pid))

                svc_list = list(filter(lambda v: not v['_remove'], svc_list))

                # actual reload of services
                for pid in pid_to_svc.keys():
                    svc = pid_to_svc[pid]

                    if svc['tune'].get('reloadable', False):
                        self._info('Reloading "{0}:{1}" pid "{2}"'.format(
                            svc['name'], svc['instanceId'], pid))
                        try:
                            svc['toolImpl'].onReload(
                                self._config, pid, svc['tune'])
                        except OSError:
                            pass
                    else:
                        self._info('Stopping "{0}:{1}" pid "{2}"'.format(
                            svc['name'], svc['instanceId'], pid))
                        self._serviceStop(svc, svc['toolImpl'], pid)

                        del pid_to_svc[pid]
                        svc['_lastExit1'] = self.__RESTART_DELAY_THRESHOLD + 1
                        svc['_lastExit2'] = 0
                        svc['_pid'] = None

                self._reload_services = False

            # create children
            for svc in svc_list:
                pid = svc['_pid']
                if pid:
                    continue

                delay = 0

                if (svc['_lastExit1'] - svc['_lastExit2']) < self.__RESTART_DELAY_THRESHOLD:
                    delay = self.__RESTART_DELAY

                pid = os.fork()

                if pid:
                    svc['_pid'] = pid
                    pid_to_svc[pid] = svc

                    if delay:
                        self._warn('Delaying start "{0}:{1}" pid "{2}"'.format(
                            svc['name'], svc['instanceId'], pid))
                    else:
                        self._info('Started "{0}:{1}" pid "{2}"'.format(
                            svc['name'], svc['instanceId'], pid))
                else:
                    try:
                        signal.signal(signal.SIGTERM, signal.SIG_DFL)
                        signal.signal(signal.SIGINT, signal.SIG_DFL)
                        signal.signal(signal.SIGHUP, signal.SIG_DFL)
                        signal.signal(signal.SIGUSR1, signal.SIG_DFL)
                        signal.signal(signal.SIGUSR2, signal.SIG_DFL)
                        signal.signal(signal.SIGCHLD, signal.SIG_DFL)

                        os.dup2(os.open(os.devnull, os.O_RDONLY), 0)

                        if delay:
                            time.sleep(delay)

                        os.chdir(self._config['wcDir'])

                        svc['toolImpl'].onRun(self._config, svc, [])
                    except Exception as e:
                        self._warn(str(e))
                        import traceback
                        self._warn(str(traceback.format_exc()))
                    finally:
                        sys.stdout.flush()
                        sys.stderr.flush()
                        # Should not be reachable here
                        os._exit(2)

            # Wait for children to exit
            try:
                self._interruptable = True

                if self._reload_services:
                    continue

                if not self._running:
                    break

                (pid, excode) = os.wait()
            except KeyboardInterrupt:
                continue
            except OSError as e:  # macOS
                self._warn(str(e))
                continue
            finally:
                self._interruptable = False

            svc = pid_to_svc[pid]
            del pid_to_svc[pid]
            svc['_pid'] = None
            svc['_lastExit2'] = svc['_lastExit1']

            try:
                svc['_lastExit1'] = time.monotonic()
            except:
                times = os.times()
                svc['_lastExit1'] = times[4]

            self._warn('Exited "{0}:{1}" pid "{2}" exit code "{3}"'.format(
                svc['name'], svc['instanceId'], pid, excode))

        # try terminate children
        #---
        self._info('Terminating children')

        for svc in svc_list:
            pid = svc['_pid']

            if pid:
                try:
                    self._info('Terminating "{0}:{1}" pid "{2}"'.format(
                        svc['name'], svc['instanceId'], pid))
                    os.kill(pid, signal.SIGTERM)
                except OSError:
                    del pid_to_svc[pid]
                    self._info('Exited "{0}:{1}" pid "{2}"'.format(
                        svc['name'], svc['instanceId'], pid))

        # try wait children
        #---
        self._info('Waiting for children')
        signal.signal(signal.SIGALRM, TimeoutException.alarmHandler)

        try:
            signal.setitimer(signal.ITIMER_REAL,
                             RuntimeTool.DEFAULT_EXIT_TIMEOUT / 1000.0)

            while len(pid_to_svc) > 0:
                (pid, excode) = os.waitpid(-1, 0)
                svc = pid_to_svc[pid]
                del pid_to_svc[pid]
                self._info('Exited "{0}:{1}" pid "{2}"'.format(
                    svc['name'], svc['instanceId'], pid))
        except TimeoutException:
            self._warn('Timed out waiting for children shutdown')
        except OSError as e:  # macOS
            self._warn(str(e))
        finally:
            signal.setitimer(signal.ITIMER_REAL, 0)

        # final kill
        #---
        if pid_to_svc:
            self._info('Killing children')
        for pid in pid_to_svc:
            try:
                svc = pid_to_svc[pid]
                self._info('Killing "{0}:{1}" pid "{2}"'.format(
                    svc['name'], svc['instanceId'], pid))
                os.kill(pid, signal.SIGKILL)
                os.waitpid(pid, 0)
            except OSError as e:
                self._warn(str(e))

        self._info('Master process exit')
        self._masterUnlock()
        self._dumpResourceStats()

    def _reloadServiceMaster(self):
        pid = self._serviceMasterPID()

        if pid:
            self._os.kill(pid, self._ext.signal.SIGUSR1)

    def _dumpResourceStats(self):
        resource = self._ext.resource
        configutil = self._configutil

        stats = resource.getrusage(resource.RUSAGE_SELF)
        self._infoLabel('Max RSS: ', configutil.toMemory(
            stats.ru_maxrss * 1024))
        self._infoLabel('User Time: ', '{0}s'.format(stats.ru_utime))
        self._infoLabel('System Time: ', '{0}s'.format(stats.ru_stime))
